"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.engine.hyperparameters import deserialize as deserialize
from keras_tuner.src.engine.hyperparameters import serialize as serialize
from keras_tuner.src.engine.hyperparameters.hp_types.boolean_hp import Boolean as Boolean
from keras_tuner.src.engine.hyperparameters.hp_types.choice_hp import Choice as Choice
from keras_tuner.src.engine.hyperparameters.hp_types.fixed_hp import Fixed as Fixed
from keras_tuner.src.engine.hyperparameters.hp_types.float_hp import Float as Float
from keras_tuner.src.engine.hyperparameters.hp_types.int_hp import Int as Int
from keras_tuner.src.engine.hyperparameters.hyperparameter import HyperParameter as HyperParameter
from keras_tuner.src.engine.hyperparameters.hyperparameters import HyperParameters as HyperParameters



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.tuners.bayesian import BayesianOptimizationOracle as BayesianOptimizationOracle
from keras_tuner.src.tuners.gridsearch import GridSearchOracle as GridSearchOracle
from keras_tuner.src.tuners.hyperband import HyperbandOracle as HyperbandOracle
from keras_tuner.src.tuners.randomsearch import RandomSearchOracle as RandomSearchOracle



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import keras


def _multi_backend():
    version_fn = getattr(keras, "version", None)
    return version_fn and version_fn().startswith("3.")


_MULTI_BACKEND = _multi_backend()


def multi_backend():
    """Check if multi-backend keras is enabled."""
    return _MULTI_BACKEND


def backend():
    """Check the backend framework."""
    return "tensorflow" if not multi_backend() else keras.config.backend()




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.engine.tuner import Tuner as Tuner
from keras_tuner.src.engine.tuner import maybe_distribute as maybe_distribute



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trial class."""


import hashlib
import random
import time

from keras_tuner.src import protos
from keras_tuner.src import utils
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.engine import hyperparameters as hp_module
from keras_tuner.src.engine import metrics_tracking
from keras_tuner.src.engine import stateful


@keras_tuner_export(["keras_tuner.engine.trial.TrialStatus"])
class TrialStatus:
    # The Trial may start to run.
    RUNNING = "RUNNING"
    # The Trial is empty. The Oracle is waiting on something else before
    # creating the trial. Should call Oracle.create_trial() again.
    IDLE = "IDLE"
    # The Trial has crashed or been deemed infeasible for the current run, but
    # subject to retries.
    INVALID = "INVALID"
    # The Trial is empty. Oracle finished searching. No new trial needed. The
    # tuner should also end the search.
    STOPPED = "STOPPED"
    # The Trial finished normally.
    COMPLETED = "COMPLETED"
    # The Trial is failed. No more retries needed.
    FAILED = "FAILED"

    @staticmethod
    def to_proto(status):
        ts = protos.get_proto().TrialStatus
        if status is None:
            return ts.UNKNOWN
        elif status == TrialStatus.RUNNING:
            return ts.RUNNING
        elif status == TrialStatus.IDLE:
            return ts.IDLE
        elif status == TrialStatus.INVALID:
            return ts.INVALID
        elif status == TrialStatus.STOPPED:
            return ts.STOPPED
        elif status == TrialStatus.COMPLETED:
            return ts.COMPLETED
        elif status == TrialStatus.FAILED:
            return ts.FAILED
        else:
            raise ValueError(f"Unknown status {status}")

    @staticmethod
    def from_proto(proto):
        ts = protos.get_proto().TrialStatus
        if proto == ts.UNKNOWN:
            return None
        elif proto == ts.RUNNING:
            return TrialStatus.RUNNING
        elif proto == ts.IDLE:
            return TrialStatus.IDLE
        elif proto == ts.INVALID:
            return TrialStatus.INVALID
        elif proto == ts.STOPPED:
            return TrialStatus.STOPPED
        elif proto == ts.COMPLETED:
            return TrialStatus.COMPLETED
        elif proto == ts.FAILED:
            return TrialStatus.FAILED
        else:
            raise ValueError(f"Unknown status {proto}")


@keras_tuner_export(["keras_tuner.engine.trial.Trial"])
class Trial(stateful.Stateful):
    """The runs with the same set of hyperparameter values.

    `Trial` objects are managed by the `Oracle`. A `Trial` object contains all
    the information related to the executions with the same set of
    hyperparameter values. A `Trial` may be executed multiple times for more
    accurate results or for retrying when failed. The related information
    includes hyperparameter values, the Trial ID, and the trial results.

    Args:
        hyperparameters: HyperParameters. It contains the hyperparameter values
            for the trial.
        trial_id: String. The unique identifier for a trial.
        status: one of the TrialStatus attributes. It marks the current status
            of the Trial.
        message: String. The error message if the trial status is "INVALID".
    """

    def __init__(
        self,
        hyperparameters,
        trial_id=None,
        status=TrialStatus.RUNNING,
        message=None,
    ):
        self.hyperparameters = hyperparameters
        self.trial_id = generate_trial_id() if trial_id is None else trial_id

        self.metrics = metrics_tracking.MetricsTracker()
        self.score = None
        self.best_step = 0
        self.status = status
        self.message = message

    def summary(self):
        """Displays a summary of this Trial."""
        print(f"Trial {self.trial_id} summary")

        print("Hyperparameters:")
        self.display_hyperparameters()

        if self.score is not None:
            print(f"Score: {self.score}")

        if self.message is not None:
            print(self.message)

    def display_hyperparameters(self):
        if self.hyperparameters.values:
            for hp, value in self.hyperparameters.values.items():
                print(f"{hp}:", value)
        else:
            print("default configuration")

    def get_state(self):
        return {
            "trial_id": self.trial_id,
            "hyperparameters": self.hyperparameters.get_config(),
            "metrics": self.metrics.get_config(),
            "score": self.score,
            "best_step": self.best_step,
            "status": self.status,
            "message": self.message,
        }

    def set_state(self, state):
        self.trial_id = state["trial_id"]
        hp = hp_module.HyperParameters.from_config(state["hyperparameters"])
        self.hyperparameters = hp
        self.metrics = metrics_tracking.MetricsTracker.from_config(
            state["metrics"]
        )
        self.score = state["score"]
        self.best_step = state["best_step"]
        self.status = state["status"]
        self.message = state["message"]

    @classmethod
    def from_state(cls, state):
        trial = cls(hyperparameters=None)
        trial.set_state(state)
        return trial

    @classmethod
    def load(cls, fname):
        return cls.from_state(utils.load_json(fname))

    def to_proto(self):
        if self.score is not None:
            score = protos.get_proto().Trial.Score(
                value=self.score, step=self.best_step
            )
        else:
            score = None
        proto = protos.get_proto().Trial(
            trial_id=self.trial_id,
            hyperparameters=self.hyperparameters.to_proto(),
            score=score,
            status=TrialStatus.to_proto(self.status),
            metrics=self.metrics.to_proto(),
        )
        return proto

    @classmethod
    def from_proto(cls, proto):
        instance = cls(
            hp_module.HyperParameters.from_proto(proto.hyperparameters),
            trial_id=proto.trial_id,
            status=TrialStatus.from_proto(proto.status),
        )
        if proto.HasField("score"):
            instance.score = proto.score.value
            instance.best_step = proto.score.step
        instance.metrics = metrics_tracking.MetricsTracker.from_proto(
            proto.metrics
        )
        return instance


def generate_trial_id():
    s = str(time.time()) + str(random.randint(1, int(1e7)))
    return hashlib.sha256(s.encode("utf-8")).hexdigest()[:32]




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keras_tuner.src import applications
from keras_tuner.src import oracles
from keras_tuner.src import tuners
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.engine.hypermodel import HyperModel
from keras_tuner.src.engine.hyperparameters import HyperParameter
from keras_tuner.src.engine.hyperparameters import HyperParameters
from keras_tuner.src.engine.objective import Objective
from keras_tuner.src.engine.oracle import Oracle
from keras_tuner.src.engine.oracle import synchronized
from keras_tuner.src.engine.tuner import Tuner
from keras_tuner.src.tuners import BayesianOptimization
from keras_tuner.src.tuners import GridSearch
from keras_tuner.src.tuners import Hyperband
from keras_tuner.src.tuners import RandomSearch
from keras_tuner.src.tuners import SklearnTuner
from keras_tuner.src.version import __version__




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hypertunable version of EfficientNet based on Keras.applications."""

from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.backend import keras
from keras_tuner.src.backend.keras import layers
from keras_tuner.src.engine import hypermodel

efficientnet = keras.applications.efficientnet
EFFICIENTNET_MODELS = {
    "B0": efficientnet.EfficientNetB0,
    "B1": efficientnet.EfficientNetB1,
    "B2": efficientnet.EfficientNetB2,
    "B3": efficientnet.EfficientNetB3,
    "B4": efficientnet.EfficientNetB4,
    "B5": efficientnet.EfficientNetB5,
    "B6": efficientnet.EfficientNetB6,
    "B7": efficientnet.EfficientNetB7,
}

EFFICIENTNET_IMG_SIZE = {
    "B0": 224,
    "B1": 240,
    "B2": 260,
    "B3": 300,
    "B4": 380,
    "B5": 456,
    "B6": 528,
    "B7": 600,
}


@keras_tuner_export("keras_tuner.applications.HyperEfficientNet")
class HyperEfficientNet(hypermodel.HyperModel):
    """An EfficientNet hypermodel.

    Models built by `HyperEfficientNet` take images with shape (height, width,
    channels) as input. The output are one-hot encoded with the length matching
    the number of classes specified by the `classes` argument.

    Args:
        input_shape: Optional shape tuple, e.g. `(256, 256, 3)`.  One of
            `input_shape` or `input_tensor` must be specified.
        input_tensor: Optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.  One of `input_shape` or
            `input_tensor` must be specified.
        classes: Optional number of classes to classify images into, only to be
            specified if `include_top` is True, and if no `weights` argument is
            specified.
        augmentation_model: Optional `Model` or `HyperModel` instance for image
            augmentation.
        **kwargs: Additional keyword arguments that apply to all hypermodels.
            See `keras_tuner.HyperModel`.
    """

    def __init__(
        self,
        input_shape=None,
        input_tensor=None,
        classes=None,
        augmentation_model=None,
        **kwargs,
    ):
        if not isinstance(
            augmentation_model, (hypermodel.HyperModel, keras.Model, type(None))
        ):
            raise ValueError(
                "Keyword augmentation_model should be "
                "a `HyperModel`, a Keras `Model` or "
                f"empty. Received {augmentation_model}."
            )

        if not classes:
            raise ValueError("You must specify `classes`.")

        if input_shape is None and input_tensor is None:
            raise ValueError(
                "You must specify either `input_shape` or `input_tensor`."
            )

        self.input_shape = input_shape
        self.input_tensor = input_tensor
        self.classes = classes
        self.augmentation_model = augmentation_model

        super().__init__(**kwargs)

    def build(self, hp):
        if self.input_tensor is not None:
            inputs = keras.utils.get_source_inputs(self.input_tensor)
            x = self.input_tensor
        else:
            inputs = layers.Input(shape=self.input_shape)
            x = inputs

        if self.augmentation_model:
            if isinstance(self.augmentation_model, hypermodel.HyperModel):
                augmentation_model = self.augmentation_model.build(hp)
            elif isinstance(self.augmentation_model, keras.models.Model):
                augmentation_model = self.augmentation_model

            x = augmentation_model(x)

        # Select one of pre-trained EfficientNet as feature extractor
        version = hp.Choice(
            "version", [f"B{i}" for i in range(8)], default="B0"
        )
        img_size = EFFICIENTNET_IMG_SIZE[version]

        x = layers.Resizing(img_size, img_size, interpolation="bilinear")(x)
        efficientnet_model = EFFICIENTNET_MODELS[version](
            include_top=False, input_tensor=x
        )

        # Rebuild top layers of the model.
        x = efficientnet_model.output

        pooling = hp.Choice("pooling", ["avg", "max"], default="avg")
        if pooling == "avg":
            x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
        elif pooling == "max":
            x = layers.GlobalMaxPooling2D(name="max_pool")(x)

        top_dropout_rate = hp.Float(
            "top_dropout_rate",
            min_value=0.2,
            max_value=0.8,
            step=0.2,
            default=0.2,
        )
        x = layers.Dropout(top_dropout_rate, name="top_dropout")(x)

        x = layers.Dense(self.classes, activation="softmax", name="probs")(x)

        # compile
        model = keras.Model(inputs, x, name="EfficientNet")
        self._compile(model, hp)

        return model

    def _compile(self, model, hp):
        """Compile model using hyperparameters in hp.

        When subclassing the hypermodel, this may be overridden to change
        behavior of compiling.
        """
        learning_rate = hp.Choice(
            "learning_rate", [0.1, 0.01, 0.001], default=0.01
        )
        optimizer = keras.optimizers.SGD(
            momentum=0.1, learning_rate=learning_rate
        )

        model.compile(
            optimizer=optimizer,
            loss="categorical_crossentropy",
            metrics=["accuracy"],
        )




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""KerasTuner utilities."""
import json

from keras_tuner.src import backend
from keras_tuner.src.backend import keras

# Check if we are in a ipython/colab environement
try:
    import IPython

    class_name = IPython.get_ipython().__class__.__name__
    IS_NOTEBOOK = "Terminal" not in class_name
except (NameError, ImportError):  # pragma: no cover
    IS_NOTEBOOK = False  # pragma: no cover


if IS_NOTEBOOK:
    from IPython import display


def try_clear():
    if IS_NOTEBOOK:
        display.clear_output()
    else:
        print()


def create_directory(path, remove_existing=False):
    # Create the directory if it doesn't exist.
    if not backend.io.exists(path):
        backend.io.makedirs(path)

    # If it does exist, and remove_existing is specified,
    # the directory will be removed and recreated.
    elif remove_existing:
        backend.io.rmtree(path)
        backend.io.makedirs(path)


def serialize_keras_object(obj):
    if hasattr(keras.utils, "legacy"):
        return keras.utils.legacy.serialize_keras_object(  # pragma: no cover
            obj
        )
    else:
        return keras.utils.serialize_keras_object(obj)  # pragma: no cover


def deserialize_keras_object(config, module_objects=None, custom_objects=None):
    if hasattr(keras.utils, "legacy"):
        return keras.utils.legacy.deserialize_keras_object(  # pragma: no cover
            config, custom_objects, module_objects
        )
    else:
        return keras.utils.deserialize_keras_object(  # pragma: no cover
            config, custom_objects, module_objects
        )


def to_list(values):
    if isinstance(values, list):
        return values
    if isinstance(values, tuple):
        return list(values)
    return [values]


def save_json(path, obj):
    """Save Python object to a json file.

    Args:
        path: String. The path to the json file.
        obj: Object. The Python object to be saved.

    Returns:
        The json string format of the object.
    """
    obj_str = json.dumps(obj)
    with backend.io.File(path, "w") as f:
        f.write(obj_str)
    return obj_str


def load_json(path):
    """Load json from file.

    Args:
        path: String. The path to the json file.

    Returns:
        A Python object.
    """
    with backend.io.File(path, "r") as f:
        obj_str = f.read()
    return json.loads(obj_str)




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.engine import base_tuner as base_tuner
from keras_tuner.engine import hypermodel as hypermodel
from keras_tuner.engine import hyperparameters as hyperparameters
from keras_tuner.engine import metrics_tracking as metrics_tracking
from keras_tuner.engine import oracle as oracle
from keras_tuner.engine import trial as trial
from keras_tuner.engine import tuner as tuner



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from keras_tuner.src import protos
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.engine import conditions as conditions_mod
from keras_tuner.src.engine.hyperparameters import hyperparameter


@keras_tuner_export("keras_tuner.engine.hyperparameters.Fixed")
class Fixed(hyperparameter.HyperParameter):
    """Fixed, untunable value.

    Args:
        name: A string. the name of parameter. Must be unique for each
            `HyperParameter` instance in the search space.
        value: The value to use (can be any JSON-serializable Python type).
    """

    def __init__(self, name, value, **kwargs):
        super().__init__(name=name, default=value, **kwargs)
        self.name = name

        if isinstance(value, bool):
            value = bool(value)
        elif isinstance(value, six.integer_types):
            value = int(value)
        elif isinstance(value, six.string_types):
            value = str(value)
        elif not isinstance(value, (float, str)):
            raise ValueError(
                "`Fixed` value must be an `int`, `float`, `str`, "
                f"or `bool`, found {value}"
            )
        self.value = value

    def __repr__(self):
        return f"Fixed(name: {self.name}, value: {self.value})"

    @property
    def values(self):
        return (self.value,)

    def prob_to_value(self, prob):
        return self.value

    def value_to_prob(self, value):
        return 0.5

    @property
    def default(self):
        return self.value

    def get_config(self):
        config = super().get_config()
        config["name"] = self.name
        config.pop("default")
        config["value"] = self.value
        return config

    @classmethod
    def from_proto(cls, proto):
        value = getattr(proto.value, proto.value.WhichOneof("kind"))
        conditions = [
            conditions_mod.Condition.from_proto(c) for c in proto.conditions
        ]
        return cls(name=proto.name, value=value, conditions=conditions)

    def to_proto(self):
        if isinstance(self.value, bool):
            # Check bool first as bool is subclass of int.
            # So bool is also six.integer_types.
            value = protos.get_proto().Value(boolean_value=self.value)
        elif isinstance(self.value, six.integer_types):
            value = protos.get_proto().Value(int_value=self.value)
        elif isinstance(self.value, float):
            value = protos.get_proto().Value(float_value=self.value)
        elif isinstance(self.value, six.string_types):
            value = protos.get_proto().Value(string_value=self.value)

        return protos.get_proto().Fixed(
            name=self.name,
            value=value,
            conditions=[c.to_proto() for c in self.conditions],
        )




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Keras backend module.

This module adds a temporarily Keras API surface that is fully under KerasTuner
control. This allows us to switch between Keras 3 and `tf.keras`, as well
as add shims to support older version of `tf.keras`.

- `config`: check which backend is being run.
- `keras`: The full `keras` API (via `keras` 3 or `tf.keras`).
- `ops`: `keras.ops`, always tf backed if using `tf.keras`.
"""

from keras_tuner.src.backend import config
from keras_tuner.src.backend import io
from keras_tuner.src.backend import keras
from keras_tuner.src.backend import ops
from keras_tuner.src.backend import random




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from keras_tuner.src import protos
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.engine import conditions as conditions_mod
from keras_tuner.src.engine.hyperparameters import hp_utils
from keras_tuner.src.engine.hyperparameters import hyperparameter


@keras_tuner_export("keras_tuner.engine.hyperparameters.Choice")
class Choice(hyperparameter.HyperParameter):
    """Choice of one value among a predefined set of possible values.

    Args:
        name: A string. the name of parameter. Must be unique for each
            `HyperParameter` instance in the search space.
        values: A list of possible values. Values must be int, float,
            str, or bool. All values must be of the same type.
        ordered: Optional boolean, whether the values passed should be
            considered to have an ordering. Defaults to `True` for float/int
            values.  Must be `False` for any other values.
        default: Optional default value to return for the parameter.
            If unspecified, the default value will be:
            - None if None is one of the choices in `values`
            - The first entry in `values` otherwise.
    """

    def __init__(self, name, values, ordered=None, default=None, **kwargs):
        super().__init__(name=name, default=default, **kwargs)
        if not values:
            raise ValueError("`values` must be provided for `Choice`.")

        # Type checking.
        types = {type(v) for v in values}
        if len(types) > 1:
            raise TypeError(
                "A `Choice` can contain only one type of value, "
                f"found values: {str(values)} with types {types}."
            )

        # Standardize on str, int, float, bool.
        if isinstance(values[0], six.string_types):
            values = [str(v) for v in values]
            if default is not None:
                default = str(default)
        elif isinstance(values[0], six.integer_types):
            values = [int(v) for v in values]
            if default is not None:
                default = int(default)
        elif not isinstance(values[0], (bool, float)):
            raise TypeError(
                "A `Choice` can contain only `int`, `float`, `str`, or "
                "`bool`, found values: " + str(values) + "with "
                "types: " + str(type(values[0]))
            )
        self._values = values

        if default is not None and default not in values:
            raise ValueError(
                "The default value should be one of the choices. "
                f"You passed: values={values}, default={default}"
            )
        self._default = default

        # Get or infer ordered.
        self.ordered = ordered
        is_numeric = isinstance(values[0], (six.integer_types, float))
        if self.ordered and not is_numeric:
            raise ValueError("`ordered` must be `False` for non-numeric types.")
        if self.ordered is None:
            self.ordered = is_numeric

    def __repr__(self):
        return (
            f"Choice(name: '{self.name}', "
            + f"values: {self._values}, "
            + f"ordered: {self.ordered}, default: {self.default})"
        )

    @property
    def values(self):
        return self._values

    @property
    def default(self):
        return self._values[0] if self._default is None else self._default

    def prob_to_value(self, prob):
        return self._values[hp_utils.prob_to_index(prob, len(self._values))]

    def value_to_prob(self, value):
        return hp_utils.index_to_prob(
            self._values.index(value), len(self._values)
        )

    def get_config(self):
        config = super().get_config()
        config["values"] = self._values
        config["ordered"] = self.ordered
        return config

    @classmethod
    def from_proto(cls, proto):
        values = [getattr(val, val.WhichOneof("kind")) for val in proto.values]
        default = getattr(proto.default, proto.default.WhichOneof("kind"), None)
        conditions = [
            conditions_mod.Condition.from_proto(c) for c in proto.conditions
        ]
        return cls(
            name=proto.name,
            values=values,
            ordered=proto.ordered,
            default=default,
            conditions=conditions,
        )

    def to_proto(self):
        if isinstance(self.values[0], six.string_types):
            values = [
                protos.get_proto().Value(string_value=v) for v in self.values
            ]
            default = protos.get_proto().Value(string_value=self.default)
        elif isinstance(self.values[0], six.integer_types):
            values = [
                protos.get_proto().Value(int_value=v) for v in self.values
            ]
            default = protos.get_proto().Value(int_value=self.default)
        else:
            values = [
                protos.get_proto().Value(float_value=v) for v in self.values
            ]
            default = protos.get_proto().Value(float_value=self.default)
        return protos.get_proto().Choice(
            name=self.name,
            ordered=self.ordered,
            values=values,
            default=default,
            conditions=[c.to_proto() for c in self.conditions],
        )




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keras_tuner/protos/keras_tuner.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$keras_tuner/protos/keras_tuner.proto\x12\x0bkeras_tuner"l\n\x05Value\x12\x13\n\tint_value\x18\x01 \x01(\x12H\x00\x12\x15\n\x0b\x66loat_value\x18\x02 \x01(\x01H\x00\x12\x16\n\x0cstring_value\x18\x03 \x01(\tH\x00\x12\x17\n\rboolean_value\x18\x04 \x01(\x08H\x00\x42\x06\n\x04kind"\xaf\x01\n\x05\x46loat\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tmin_value\x18\x02 \x01(\x01\x12\x11\n\tmax_value\x18\x03 \x01(\x01\x12\x0c\n\x04step\x18\x04 \x01(\x01\x12\'\n\x08sampling\x18\x05 \x01(\x0e\x32\x15.keras_tuner.Sampling\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x06 \x01(\x01\x12*\n\nconditions\x18\x07 \x03(\x0b\x32\x16.keras_tuner.Condition"\xad\x01\n\x03Int\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tmin_value\x18\x02 \x01(\x12\x12\x11\n\tmax_value\x18\x03 \x01(\x12\x12\x0c\n\x04step\x18\x04 \x01(\x12\x12\'\n\x08sampling\x18\x05 \x01(\x0e\x32\x15.keras_tuner.Sampling\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x06 \x01(\x12\x12*\n\nconditions\x18\x07 \x03(\x0b\x32\x16.keras_tuner.Condition"\x9c\x01\n\x06\x43hoice\x12\x0c\n\x04name\x18\x01 \x01(\t\x12"\n\x06values\x18\x02 \x03(\x0b\x32\x12.keras_tuner.Value\x12#\n\x07\x64\x65\x66\x61ult\x18\x03 \x01(\x0b\x32\x12.keras_tuner.Value\x12\x0f\n\x07ordered\x18\x04 \x01(\x08\x12*\n\nconditions\x18\x05 \x03(\x0b\x32\x16.keras_tuner.Condition"T\n\x07\x42oolean\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x02 \x01(\x08\x12*\n\nconditions\x18\x03 \x03(\x0b\x32\x16.keras_tuner.Condition"d\n\x05\x46ixed\x12\x0c\n\x04name\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.keras_tuner.Value\x12*\n\nconditions\x18\x03 \x03(\x0b\x32\x16.keras_tuner.Condition"\xe1\x03\n\x0fHyperParameters\x12\x31\n\x05space\x18\x01 \x01(\x0b\x32".keras_tuner.HyperParameters.Space\x12\x33\n\x06values\x18\x02 \x01(\x0b\x32#.keras_tuner.HyperParameters.Values\x1a\xd6\x01\n\x05Space\x12\'\n\x0b\x66loat_space\x18\x01 \x03(\x0b\x32\x12.keras_tuner.Float\x12#\n\tint_space\x18\x02 \x03(\x0b\x32\x10.keras_tuner.Int\x12)\n\x0c\x63hoice_space\x18\x03 \x03(\x0b\x32\x13.keras_tuner.Choice\x12+\n\rboolean_space\x18\x04 \x03(\x0b\x32\x14.keras_tuner.Boolean\x12\'\n\x0b\x66ixed_space\x18\x05 \x03(\x0b\x32\x12.keras_tuner.Fixed\x1a\x8c\x01\n\x06Values\x12?\n\x06values\x18\x01 \x03(\x0b\x32/.keras_tuner.HyperParameters.Values.ValuesEntry\x1a\x41\n\x0bValuesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.keras_tuner.Value:\x02\x38\x01"0\n\x11MetricObservation\x12\r\n\x05value\x18\x01 \x03(\x02\x12\x0c\n\x04step\x18\x02 \x01(\x03"W\n\rMetricHistory\x12\x34\n\x0cobservations\x18\x01 \x03(\x0b\x32\x1e.keras_tuner.MetricObservation\x12\x10\n\x08maximize\x18\x02 \x01(\x08"\x97\x01\n\x0eMetricsTracker\x12\x39\n\x07metrics\x18\x01 \x03(\x0b\x32(.keras_tuner.MetricsTracker.MetricsEntry\x1aJ\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.keras_tuner.MetricHistory:\x02\x38\x01"\xf7\x01\n\x05Trial\x12\x35\n\x0fhyperparameters\x18\x01 \x01(\x0b\x32\x1c.keras_tuner.HyperParameters\x12\x10\n\x08trial_id\x18\x02 \x01(\t\x12(\n\x06status\x18\x03 \x01(\x0e\x32\x18.keras_tuner.TrialStatus\x12,\n\x07metrics\x18\x04 \x01(\x0b\x32\x1b.keras_tuner.MetricsTracker\x12\'\n\x05score\x18\x05 \x01(\x0b\x32\x18.keras_tuner.Trial.Score\x1a$\n\x05Score\x12\r\n\x05value\x18\x01 \x01(\x02\x12\x0c\n\x04step\x18\x02 \x01(\x03"\x80\x01\n\tCondition\x12/\n\x06parent\x18\x01 \x01(\x0b\x32\x1d.keras_tuner.Condition.ParentH\x00\x1a:\n\x06Parent\x12\x0c\n\x04name\x18\x01 \x01(\t\x12"\n\x06values\x18\x02 \x03(\x0b\x32\x12.keras_tuner.ValueB\x06\n\x04kind*:\n\x08Sampling\x12\x08\n\x04NONE\x10\x00\x12\n\n\x06LINEAR\x10\x01\x12\x07\n\x03LOG\x10\x02\x12\x0f\n\x0bREVERSE_LOG\x10\x03*f\n\x0bTrialStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\x08\n\x04IDLE\x10\x02\x12\x0b\n\x07INVALID\x10\x03\x12\x0b\n\x07STOPPED\x10\x04\x12\r\n\tCOMPLETED\x10\x05\x12\n\n\x06\x46\x41ILED\x10\x06\x62\x06proto3'
)

_SAMPLING = DESCRIPTOR.enum_types_by_name["Sampling"]
Sampling = enum_type_wrapper.EnumTypeWrapper(_SAMPLING)
_TRIALSTATUS = DESCRIPTOR.enum_types_by_name["TrialStatus"]
TrialStatus = enum_type_wrapper.EnumTypeWrapper(_TRIALSTATUS)
NONE = 0
LINEAR = 1
LOG = 2
REVERSE_LOG = 3
UNKNOWN = 0
RUNNING = 1
IDLE = 2
INVALID = 3
STOPPED = 4
COMPLETED = 5
FAILED = 6


_VALUE = DESCRIPTOR.message_types_by_name["Value"]
_FLOAT = DESCRIPTOR.message_types_by_name["Float"]
_INT = DESCRIPTOR.message_types_by_name["Int"]
_CHOICE = DESCRIPTOR.message_types_by_name["Choice"]
_BOOLEAN = DESCRIPTOR.message_types_by_name["Boolean"]
_FIXED = DESCRIPTOR.message_types_by_name["Fixed"]
_HYPERPARAMETERS = DESCRIPTOR.message_types_by_name["HyperParameters"]
_HYPERPARAMETERS_SPACE = _HYPERPARAMETERS.nested_types_by_name["Space"]
_HYPERPARAMETERS_VALUES = _HYPERPARAMETERS.nested_types_by_name["Values"]
_HYPERPARAMETERS_VALUES_VALUESENTRY = (
    _HYPERPARAMETERS_VALUES.nested_types_by_name["ValuesEntry"]
)
_METRICOBSERVATION = DESCRIPTOR.message_types_by_name["MetricObservation"]
_METRICHISTORY = DESCRIPTOR.message_types_by_name["MetricHistory"]
_METRICSTRACKER = DESCRIPTOR.message_types_by_name["MetricsTracker"]
_METRICSTRACKER_METRICSENTRY = _METRICSTRACKER.nested_types_by_name[
    "MetricsEntry"
]
_TRIAL = DESCRIPTOR.message_types_by_name["Trial"]
_TRIAL_SCORE = _TRIAL.nested_types_by_name["Score"]
_CONDITION = DESCRIPTOR.message_types_by_name["Condition"]
_CONDITION_PARENT = _CONDITION.nested_types_by_name["Parent"]
Value = _reflection.GeneratedProtocolMessageType(
    "Value",
    (_message.Message,),
    {
        "DESCRIPTOR": _VALUE,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Value)
    },
)
_sym_db.RegisterMessage(Value)

Float = _reflection.GeneratedProtocolMessageType(
    "Float",
    (_message.Message,),
    {
        "DESCRIPTOR": _FLOAT,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Float)
    },
)
_sym_db.RegisterMessage(Float)

Int = _reflection.GeneratedProtocolMessageType(
    "Int",
    (_message.Message,),
    {
        "DESCRIPTOR": _INT,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Int)
    },
)
_sym_db.RegisterMessage(Int)

Choice = _reflection.GeneratedProtocolMessageType(
    "Choice",
    (_message.Message,),
    {
        "DESCRIPTOR": _CHOICE,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Choice)
    },
)
_sym_db.RegisterMessage(Choice)

Boolean = _reflection.GeneratedProtocolMessageType(
    "Boolean",
    (_message.Message,),
    {
        "DESCRIPTOR": _BOOLEAN,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Boolean)
    },
)
_sym_db.RegisterMessage(Boolean)

Fixed = _reflection.GeneratedProtocolMessageType(
    "Fixed",
    (_message.Message,),
    {
        "DESCRIPTOR": _FIXED,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Fixed)
    },
)
_sym_db.RegisterMessage(Fixed)

HyperParameters = _reflection.GeneratedProtocolMessageType(
    "HyperParameters",
    (_message.Message,),
    {
        "Space": _reflection.GeneratedProtocolMessageType(
            "Space",
            (_message.Message,),
            {
                "DESCRIPTOR": _HYPERPARAMETERS_SPACE,
                "__module__": "keras_tuner.protos.keras_tuner_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.HyperParameters.Space)
            },
        ),
        "Values": _reflection.GeneratedProtocolMessageType(
            "Values",
            (_message.Message,),
            {
                "ValuesEntry": _reflection.GeneratedProtocolMessageType(
                    "ValuesEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _HYPERPARAMETERS_VALUES_VALUESENTRY,
                        "__module__": "keras_tuner.protos.keras_tuner_pb2",
                        # @@protoc_insertion_point(class_scope:keras_tuner.HyperParameters.Values.ValuesEntry)
                    },
                ),
                "DESCRIPTOR": _HYPERPARAMETERS_VALUES,
                "__module__": "keras_tuner.protos.keras_tuner_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.HyperParameters.Values)
            },
        ),
        "DESCRIPTOR": _HYPERPARAMETERS,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.HyperParameters)
    },
)
_sym_db.RegisterMessage(HyperParameters)
_sym_db.RegisterMessage(HyperParameters.Space)
_sym_db.RegisterMessage(HyperParameters.Values)
_sym_db.RegisterMessage(HyperParameters.Values.ValuesEntry)

MetricObservation = _reflection.GeneratedProtocolMessageType(
    "MetricObservation",
    (_message.Message,),
    {
        "DESCRIPTOR": _METRICOBSERVATION,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.MetricObservation)
    },
)
_sym_db.RegisterMessage(MetricObservation)

MetricHistory = _reflection.GeneratedProtocolMessageType(
    "MetricHistory",
    (_message.Message,),
    {
        "DESCRIPTOR": _METRICHISTORY,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.MetricHistory)
    },
)
_sym_db.RegisterMessage(MetricHistory)

MetricsTracker = _reflection.GeneratedProtocolMessageType(
    "MetricsTracker",
    (_message.Message,),
    {
        "MetricsEntry": _reflection.GeneratedProtocolMessageType(
            "MetricsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _METRICSTRACKER_METRICSENTRY,
                "__module__": "keras_tuner.protos.keras_tuner_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.MetricsTracker.MetricsEntry)
            },
        ),
        "DESCRIPTOR": _METRICSTRACKER,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.MetricsTracker)
    },
)
_sym_db.RegisterMessage(MetricsTracker)
_sym_db.RegisterMessage(MetricsTracker.MetricsEntry)

Trial = _reflection.GeneratedProtocolMessageType(
    "Trial",
    (_message.Message,),
    {
        "Score": _reflection.GeneratedProtocolMessageType(
            "Score",
            (_message.Message,),
            {
                "DESCRIPTOR": _TRIAL_SCORE,
                "__module__": "keras_tuner.protos.keras_tuner_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.Trial.Score)
            },
        ),
        "DESCRIPTOR": _TRIAL,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Trial)
    },
)
_sym_db.RegisterMessage(Trial)
_sym_db.RegisterMessage(Trial.Score)

Condition = _reflection.GeneratedProtocolMessageType(
    "Condition",
    (_message.Message,),
    {
        "Parent": _reflection.GeneratedProtocolMessageType(
            "Parent",
            (_message.Message,),
            {
                "DESCRIPTOR": _CONDITION_PARENT,
                "__module__": "keras_tuner.protos.keras_tuner_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.Condition.Parent)
            },
        ),
        "DESCRIPTOR": _CONDITION,
        "__module__": "keras_tuner.protos.keras_tuner_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.Condition)
    },
)
_sym_db.RegisterMessage(Condition)
_sym_db.RegisterMessage(Condition.Parent)

if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _HYPERPARAMETERS_VALUES_VALUESENTRY._options = None
    _HYPERPARAMETERS_VALUES_VALUESENTRY._serialized_options = b"8\001"
    _METRICSTRACKER_METRICSENTRY._options = None
    _METRICSTRACKER_METRICSENTRY._serialized_options = b"8\001"
    _SAMPLING._serialized_start = 2022
    _SAMPLING._serialized_end = 2080
    _TRIALSTATUS._serialized_start = 2082
    _TRIALSTATUS._serialized_end = 2184
    _VALUE._serialized_start = 53
    _VALUE._serialized_end = 161
    _FLOAT._serialized_start = 164
    _FLOAT._serialized_end = 339
    _INT._serialized_start = 342
    _INT._serialized_end = 515
    _CHOICE._serialized_start = 518
    _CHOICE._serialized_end = 674
    _BOOLEAN._serialized_start = 676
    _BOOLEAN._serialized_end = 760
    _FIXED._serialized_start = 762
    _FIXED._serialized_end = 862
    _HYPERPARAMETERS._serialized_start = 865
    _HYPERPARAMETERS._serialized_end = 1346
    _HYPERPARAMETERS_SPACE._serialized_start = 989
    _HYPERPARAMETERS_SPACE._serialized_end = 1203
    _HYPERPARAMETERS_VALUES._serialized_start = 1206
    _HYPERPARAMETERS_VALUES._serialized_end = 1346
    _HYPERPARAMETERS_VALUES_VALUESENTRY._serialized_start = 1281
    _HYPERPARAMETERS_VALUES_VALUESENTRY._serialized_end = 1346
    _METRICOBSERVATION._serialized_start = 1348
    _METRICOBSERVATION._serialized_end = 1396
    _METRICHISTORY._serialized_start = 1398
    _METRICHISTORY._serialized_end = 1485
    _METRICSTRACKER._serialized_start = 1488
    _METRICSTRACKER._serialized_end = 1639
    _METRICSTRACKER_METRICSENTRY._serialized_start = 1565
    _METRICSTRACKER_METRICSENTRY._serialized_end = 1639
    _TRIAL._serialized_start = 1642
    _TRIAL._serialized_end = 1889
    _TRIAL_SCORE._serialized_start = 1853
    _TRIAL_SCORE._serialized_end = 1889
    _CONDITION._serialized_start = 1892
    _CONDITION._serialized_end = 2020
    _CONDITION_PARENT._serialized_start = 1954
    _CONDITION_PARENT._serialized_end = 2012
# @@protoc_insertion_point(module_scope)




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import numpy as np
import six

from keras_tuner.src import protos
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.backend import keras


class MetricObservation:
    """Metric value at a given step of training across multiple executions.

    If the model is trained multiple
    times (multiple executions), KerasTuner records the value of each
    metric at each training step. These values are aggregated
    over multiple executions into a list where each value corresponds
    to one execution.

    Args:
        value: Float or a list of floats. The evaluated metric values.
        step: Int. The step of the evaluation, for example, the epoch number.
    """

    def __init__(self, value, step):
        if not isinstance(value, list):
            value = [value]
        self.value = value
        self.step = step

    def append(self, value):
        if not isinstance(value, list):
            value = [value]
        self.value += value

    def mean(self):
        return np.mean(self.value)

    def get_config(self):
        return {"value": self.value, "step": self.step}

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def __eq__(self, other):
        return (
            other.value == self.value and other.step == self.step
            if isinstance(other, MetricObservation)
            else False
        )

    def __repr__(self):
        return f"MetricObservation(value={self.value}, step={self.step})"

    def to_proto(self):
        return protos.get_proto().MetricObservation(
            value=self.value, step=self.step
        )

    @classmethod
    def from_proto(cls, proto):
        return cls(value=list(proto.value), step=proto.step)


class MetricHistory:
    """Record of multiple executions of a single metric.

    It contains a collection of `MetricObservation` instances.

    Args:
        direction: String. The direction of the metric to optimize. The value
            should be "min" or "max".
    """

    def __init__(self, direction="min"):
        if direction not in {"min", "max"}:
            raise ValueError(
                "`direction` should be one of "
                '{"min", "max"}, but got: %s' % (direction,)
            )
        self.direction = direction
        # Mapping step to `MetricObservation`.
        self._observations = {}

    def update(self, value, step):
        if step in self._observations:
            self._observations[step].append(value)
        else:
            self._observations[step] = MetricObservation(value, step=step)

    def get_best_value(self):
        values = [obs.mean() for obs in self._observations.values()]
        if not values:
            return None
        return (
            np.nanmin(values) if self.direction == "min" else np.nanmax(values)
        )

    def get_best_step(self):
        best_value = self.get_best_value()
        if best_value is None:
            return None
        for obs in self._observations.values():
            if obs.mean() == best_value:
                return obs.step

    def get_history(self):
        return sorted(self._observations.values(), key=lambda obs: obs.step)

    def set_history(self, observations):
        for obs in observations:
            self.update(obs.value, step=obs.step)

    def get_statistics(self):
        history = self.get_history()
        history_values = [obs.mean() for obs in history]
        return (
            {
                "min": float(np.nanmin(history_values)),
                "max": float(np.nanmax(history_values)),
                "mean": float(np.nanmean(history_values)),
                "median": float(np.nanmedian(history_values)),
                "var": float(np.nanvar(history_values)),
                "std": float(np.nanstd(history_values)),
            }
            if len(history_values)
            else {}
        )

    def get_last_value(self):
        history = self.get_history()
        if history:
            last_obs = history[-1]
            return last_obs.mean()
        else:
            return None

    def get_config(self):
        config = {
            "direction": self.direction,
            "observations": [obs.get_config() for obs in self.get_history()],
        }

        return config

    @classmethod
    def from_config(cls, config):
        instance = cls(config["direction"])
        instance.set_history(
            [
                MetricObservation.from_config(obs)
                for obs in config["observations"]
            ]
        )
        return instance

    def to_proto(self):
        return protos.get_proto().MetricHistory(
            observations=[obs.to_proto() for obs in self.get_history()],
            maximize=self.direction == "max",
        )

    @classmethod
    def from_proto(cls, proto):
        direction = "max" if proto.maximize else "min"
        instance = cls(direction)
        instance.set_history(
            [MetricObservation.from_proto(p) for p in proto.observations]
        )
        return instance


class MetricsTracker:
    """Record of the values of multiple executions of all metrics.

    It contains `MetricHistory` instances for the metrics.

    Args:
        metrics: List of strings of the names of the metrics.
    """

    def __init__(self, metrics=None):
        # str -> MetricHistory
        self.metrics = {}
        self.register_metrics(metrics)

    def exists(self, name):
        return name in self.metrics

    def register_metrics(self, metrics=None):
        metrics = metrics or []
        for metric in metrics:
            self.register(metric.name)

    def register(self, name, direction=None):
        if self.exists(name):
            raise ValueError(f"Metric already exists: {name}")
        if direction is None:
            direction = infer_metric_direction(name)
        if direction is None:
            # Objective direction is handled separately, but
            # non-objective direction defaults to min.
            direction = "min"
        self.metrics[name] = MetricHistory(direction)

    def update(self, name, value, step=0):
        value = float(value)
        if not self.exists(name):
            self.register(name)

        prev_best = self.metrics[name].get_best_value()
        self.metrics[name].update(value, step=step)
        new_best = self.metrics[name].get_best_value()

        improved = new_best != prev_best
        return improved

    def get_history(self, name):
        self._assert_exists(name)
        return self.metrics[name].get_history()

    def set_history(self, name, observations):
        if not self.exists(name):
            self.register(name)
        self.metrics[name].set_history(observations)

    def get_best_value(self, name):
        self._assert_exists(name)
        return self.metrics[name].get_best_value()

    def get_best_step(self, name):
        self._assert_exists(name)
        return self.metrics[name].get_best_step()

    def get_statistics(self, name):
        self._assert_exists(name)
        return self.metrics[name].get_statistics()

    def get_last_value(self, name):
        self._assert_exists(name)
        return self.metrics[name].get_last_value()

    def get_direction(self, name):
        self._assert_exists(name)
        return self.metrics[name].direction

    def get_config(self):
        return {
            "metrics": {
                name: metric_history.get_config()
                for name, metric_history in self.metrics.items()
            }
        }

    @classmethod
    def from_config(cls, config):
        instance = cls()
        instance.metrics = {
            name: MetricHistory.from_config(metric_history)
            for name, metric_history in config["metrics"].items()
        }
        return instance

    def to_proto(self):
        return protos.get_proto().MetricsTracker(
            metrics={
                name: metric_history.to_proto()
                for name, metric_history in self.metrics.items()
            }
        )

    @classmethod
    def from_proto(cls, proto):
        instance = cls()
        instance.metrics = {
            name: MetricHistory.from_proto(metric_history)
            for name, metric_history in proto.metrics.items()
        }
        return instance

    def _assert_exists(self, name):
        if name not in self.metrics:
            raise ValueError(f"Unknown metric: {name}")


_MAX_METRICS = (
    "Accuracy",
    "BinaryAccuracy",
    "CategoricalAccuracy",
    "SparseCategoricalAccuracy",
    "TopKCategoricalAccuracy",
    "SparseTopKCategoricalAccuracy",
    "TruePositives",
    "TrueNegatives",
    "Precision",
    "Recall",
    "AUC",
    "SensitivityAtSpecificity",
    "SpecificityAtSensitivity",
)

_MAX_METRIC_FNS = (
    "accuracy",
    "categorical_accuracy",
    "binary_accuracy",
    "sparse_categorical_accuracy",
)


@keras_tuner_export(
    "keras_tuner.engine.metrics_tracking.infer_metric_direction",
)
def infer_metric_direction(metric):
    # Handle str input and get canonical object.
    if isinstance(metric, six.string_types):
        metric_name = metric

        if metric_name.startswith("val_"):
            metric_name = metric_name.replace("val_", "", 1)

        if metric_name.startswith("weighted_"):
            metric_name = metric_name.replace("weighted_", "", 1)

        # Special-cases (from `keras/engine/training_utils.py`)
        if metric_name in {"loss", "crossentropy", "ce"}:
            return "min"
        elif metric_name == "acc":
            return "max"

        try:
            if (
                "use_legacy_format"
                in inspect.getfullargspec(keras.metrics.deserialize).args
            ):
                metric = keras.metrics.deserialize(  # pragma: no cover
                    metric_name, use_legacy_format=True
                )
            else:
                metric = keras.metrics.deserialize(  # pragma: no cover
                    metric_name
                )
        except ValueError:
            try:
                if (
                    "use_legacy_format"
                    in inspect.getfullargspec(keras.losses.deserialize).args
                ):
                    metric = keras.losses.deserialize(  # pragma: no cover
                        metric_name, use_legacy_format=True
                    )
                else:
                    metric = keras.losses.deserialize(  # pragma: no cover
                        metric_name
                    )
            except Exception:
                # Direction can't be inferred.
                return None

    # Metric class, Loss class, or function.
    if isinstance(metric, (keras.metrics.Metric, keras.losses.Loss)):
        name = metric.__class__.__name__
        if name == "MeanMetricWrapper":
            name = metric._fn.__name__  # pragma: no cover
    elif isinstance(metric, str):
        name = metric
    else:
        name = metric.__name__

    if name in _MAX_METRICS or name in _MAX_METRIC_FNS:
        return "max"
    elif hasattr(keras.metrics, name) or hasattr(keras.losses, name):
        return "min"

    # Direction can't be inferred.
    return None




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.backend import keras
from keras_tuner.src.backend import ops
from keras_tuner.src.backend import random
from keras_tuner.src.backend.keras import layers
from keras_tuner.src.engine import hypermodel

# dict of functions that create layers for transforms.
# Each function takes a factor (0 to 1) for the strength
# of the transform.
TRANSFORMS = {
    "translate_x": lambda x: layers.RandomTranslation(x, 0),
    "translate_y": lambda y: layers.RandomTranslation(0, y),
    "rotate": layers.RandomRotation,
    "contrast": layers.RandomContrast,
}


@keras.saving.register_keras_serializable(package="keras_tuner")
class RandAugment(keras.layers.Layer):
    """A single RandAugment layer."""

    def __init__(self, layers, factors):
        super().__init__()
        self.layers = layers
        self.factors = factors

    def call(self, inputs):
        x = inputs
        batch_size = ops.shape(x)[0]
        # selection tensor determines operation for each sample.
        selection = ops.cast(
            random.uniform((batch_size, 1, 1, 1), maxval=len(self.layers)),
            dtype="int32",
        )

        for i, layer in enumerate(self.layers):
            factor = self.factors[i]
            if factor == 0:
                continue
            transform_layer = TRANSFORMS[layer](factor)
            x_trans = transform_layer(x)

            # For each sample, apply the transform if and only if
            # selection matches the transform index `i`
            x = ops.where(ops.equal(i, selection), x_trans, x)

        return x

    def compute_output_shape(self, input_shape):
        return input_shape


@keras_tuner_export("keras_tuner.applications.HyperImageAugment")
class HyperImageAugment(hypermodel.HyperModel):
    """A image augmentation hypermodel.

    The `HyperImageAugment` class searches for the best combination of image
    augmentation operations in Keras preprocessing layers. The input shape of
    the model should be (height, width, channels). The output of the model is
    of the same shape as the input.

    Args:
        input_shape: Optional shape tuple, e.g. `(256, 256, 3)`.
        input_tensor: Optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.
        rotate: A number between [0, 1], a list of two numbers between [0, 1]
            or None. Configures the search space of the factor of random
            rotation transform in the augmentation. A factor is chosen for each
            trial. It sets maximum of clockwise and counterclockwise rotation
            in terms of fraction of pi, among all samples in the trial.
            Default is 0.5. When `rotate` is a single number, the search range
            is [0, `rotate`].
            The transform is off when set to None.
        translate_x: A number between [0, 1], a list of two numbers between
            [0, 1] or None. Configures the search space of the factor of random
            horizontal translation transform in the augmentation. A factor is
            chosen for each trial. It sets maximum of horizontal translation in
            terms of ratio over the width among all samples in the trial.
            Default is 0.4. When `translate_x` is a single number, the search
            range is [0, `translate_x`].
            The transform is off when set to None.
        translate_y: A number between [0, 1], a list of two numbers between
            [0, 1] or None. Configures the search space of the factor of random
            vertical translation transform in the augmentation. A factor is
            chosen for each trial. It sets maximum of vertical translation in
            terms of ratio over the height among all samples in the trial.
            Default is 0.4. When `translate_y` is a single number ,the search
            range is [0, `translate_y`].  The transform is off when set to None.
        contrast: A number between [0, 1], a list of two numbers between [0, 1]
            or None. Configures the search space of the factor of random
            contrast transform in the augmentation. A factor is chosen for each
            trial. It sets maximum ratio of contrast change among all samples in
            the trial. Default is 0.3. When `contrast` is a single number, the
            search rnage is [0, `contrast`].
            The transform is off when set to None.
        augment_layers: None, int or list of two ints, controlling the number
            of augment applied. Default is 3.
            When `augment_layers` is 0, all transform are applied sequentially.
            When `augment_layers` is nonzero, or a list of two ints, a simple
            version of RandAugment(https://arxiv.org/abs/1909.13719) is used.
            A search space for 'augment_layers' is created to search [0,
            `augment_layers`], or between the two ints if a `augment_layers` is
            a list. For each trial, the hyperparameter 'augment_layers'
            determines number of layers of augment transforms are applied,
            each randomly picked from all available transform types with equal
            probability on each sample.
        **kwargs: Additional keyword arguments that apply to all hypermodels.
            See `keras_tuner.HyperModel`.

    Example:

    ```python
    hm_aug = HyperImageAugment(input_shape=(32, 32, 3),
                               augment_layers=0,
                               rotate=[0.2, 0.3],
                               translate_x=0.1,
                               translate_y=None,
                               contrast=None)
    ```

    Then the hypermodel `hm_aug` will search 'factor_rotate' between [0.2, 0.3]
    and 'factor_translate_x' between [0, 0.1]. These two augments are applied
    on all samples with factor picked per each trial.

    ```python
    hm_aug = HyperImageAugment(input_shape=(32, 32, 3),
                               translate_x=0.5,
                               translate_y=[0.2, 0.4]
                               contrast=None)
    ```

    Then the hypermodel `hm_aug` will search 'factor_rotate' between [0, 0.2],
    'factor_translate_x' between [0, 0.5], 'factor_translate_y' between
    [0.2, 0.4]. It will use RandAugment, searching 'augment_layers'
    between [0, 3]. Each layer on each sample will be chosen from rotate,
    translate_x and translate_y.
    """

    def __init__(
        self,
        input_shape=None,
        input_tensor=None,
        rotate=0.5,
        translate_x=0.4,
        translate_y=0.4,
        contrast=0.3,
        augment_layers=3,
        **kwargs,
    ):
        if input_shape is None and input_tensor is None:
            raise ValueError(
                "You must specify either `input_shape` or `input_tensor`."
            )

        self.transforms = []
        self._register_transform("rotate", rotate)
        self._register_transform("translate_x", translate_x)
        self._register_transform("translate_y", translate_y)
        self._register_transform("contrast", contrast)

        self.input_shape = input_shape
        self.input_tensor = input_tensor

        if augment_layers:
            self.model_name = "image_rand_augment"
            try:
                augment_layers_min = augment_layers[0]
                augment_layers_max = augment_layers[1]
            except TypeError:
                augment_layers_min = 1
                augment_layers_max = augment_layers
            if not (
                isinstance(augment_layers_min, int)
                and isinstance(augment_layers_max, int)
            ):
                raise ValueError(
                    "Keyword argument `augment_layers` must be int, "
                    f"but received {augment_layers}."
                )

            self.augment_layers_min = augment_layers_min
            self.augment_layers_max = augment_layers_max
        else:
            # Separatedly tune and apply all augment transforms if
            # `randaug_count` is set to 0.
            self.model_name = "image_augment"

        super().__init__(**kwargs)

    def build(self, hp):
        if self.input_tensor is not None:
            inputs = keras.utils.get_source_inputs(self.input_tensor)
            x = self.input_tensor
        else:
            inputs = layers.Input(shape=self.input_shape)
            x = inputs

        if self.model_name == "image_rand_augment":
            x = self._build_randaug_layers(x, hp)
        else:
            x = self._build_fixedaug_layers(x, hp)

        model = keras.Model(inputs, x, name=self.model_name)
        return model

    def _build_randaug_layers(self, inputs, hp):
        augment_layers = hp.Int(
            "augment_layers",
            self.augment_layers_min,
            self.augment_layers_max,
            default=self.augment_layers_min,
        )
        x = inputs
        for _ in range(augment_layers):
            factors = []
            for i, (transform, (f_min, f_max)) in enumerate(self.transforms):
                # Factor for each transform is determined per each trial.
                factor = hp.Float(
                    f"factor_{transform}", f_min, f_max, default=f_min
                )
                factors.append(factor)
            x = RandAugment([layer for layer, _ in self.transforms], factors)(x)
        return x

    def _build_fixedaug_layers(self, inputs, hp):
        x = inputs
        for transform, (factor_min, factor_max) in self.transforms:
            transform_factor = hp.Float(
                f"factor_{transform}",
                factor_min,
                factor_max,
                step=0.05,
                default=factor_min,
            )
            if transform_factor == 0:
                continue
            transform_layer = TRANSFORMS[transform](transform_factor)
            x = transform_layer(x)
        return x

    def _register_transform(self, transform_name, transform_params):
        """Register a transform and format parameters for tuning the transform.

        Args:
            transform_name: A string, the name of the transform.
            trnasform_params: A number between [0, 1], a list of two numbers
                between [0, 1] or None. If set to a single number x, the
                corresponding transform factor will be between [0, x].
                If set to a list of 2 numbers [x, y], the factor will be
                between [x, y]. If set to None, the transform will be excluded.
        """
        if not transform_params:
            return

        try:
            transform_factor_min = transform_params[0]
            transform_factor_max = transform_params[1]
            if len(transform_params) > 2:
                raise ValueError(
                    "Length of keyword argument "
                    f"{transform_name} must not exceed 2."
                )
        except TypeError:
            transform_factor_min = 0
            transform_factor_max = transform_params

        if not (
            isinstance(transform_factor_max, (int, float))
            and isinstance(transform_factor_min, (int, float))
        ):
            raise ValueError(
                f"Keyword argument {transform_name} must be int "
                f"or float, but received {transform_params}."
            )

        self.transforms.append(
            (transform_name, (transform_factor_min, transform_factor_max))
        )




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.errors import FailedTrialError as FailedTrialError
from keras_tuner.src.errors import FatalError as FatalError
from keras_tuner.src.errors import FatalRuntimeError as FatalRuntimeError
from keras_tuner.src.errors import FatalTypeError as FatalTypeError
from keras_tuner.src.errors import FatalValueError as FatalValueError



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_tuner.src import protos
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.engine import conditions as conditions_mod
from keras_tuner.src.engine.hyperparameters import hp_utils
from keras_tuner.src.engine.hyperparameters.hp_types import numerical


@keras_tuner_export("keras_tuner.engine.hyperparameters.Float")
class Float(numerical.Numerical):
    """Floating point value hyperparameter.

    Example #1:

    ```py
    hp.Float(
        "image_rotation_factor",
        min_value=0,
        max_value=1)
    ```

    All values in interval [0, 1] have equal probability of being sampled.

    Example #2:

    ```py
    hp.Float(
        "image_rotation_factor",
        min_value=0,
        max_value=1,
        step=0.2)
    ```

    `step` is the minimum distance between samples.
    The possible values are [0, 0.2, 0.4, 0.6, 0.8, 1.0].

    Example #3:

    ```py
    hp.Float(
        "learning_rate",
        min_value=0.001,
        max_value=10,
        step=10,
        sampling="log")
    ```

    When `sampling="log"`, the `step` is multiplied between samples.
    The possible values are [0.001, 0.01, 0.1, 1, 10].

    Args:
        name: A string. the name of parameter. Must be unique for each
            `HyperParameter` instance in the search space.
        min_value: Float, the lower bound of the range.
        max_value: Float, the upper bound of the range.
        step: Optional float, the distance between two consecutive samples in
            the range. If left unspecified, it is possible to sample any value
            in the interval. If `sampling="linear"`, it will be the minimum
            additve between two samples. If `sampling="log"`, it will be the
            minimum multiplier between two samples.
        sampling: String. One of "linear", "log", "reverse_log". Defaults to
            "linear". When sampling value, it always start from a value in range
            [0.0, 1.0). The `sampling` argument decides how the value is
            projected into the range of [min_value, max_value].
            "linear": min_value + value * (max_value - min_value)
            "log": min_value * (max_value / min_value) ^ value
            "reverse_log":
                max_value - min_value * ((max_value/min_value)^(1 - value) - 1)
        default: Float, the default value to return for the parameter. If
            unspecified, the default value will be `min_value`.
    """

    def __init__(
        self,
        name,
        min_value,
        max_value,
        step=None,
        sampling="linear",
        default=None,
        **kwargs,
    ):
        if step is not None:
            self.step = float(step)
        super().__init__(
            name=name,
            min_value=float(min_value),
            max_value=float(max_value),
            step=step,
            sampling=sampling,
            default=default,
            **kwargs,
        )

    def __repr__(self):
        return (
            f"Float(name: '{self.name}', min_value: '{self.min_value}', "
            f"max_value: '{self.max_value}', step: '{self.step}', "
            f"sampling: '{self.sampling}', default: '{self.default}')"
        )

    @property
    def default(self):
        return self._default if self._default is not None else self.min_value

    def prob_to_value(self, prob):
        if self.step is None:
            return self._sample_numerical_value(prob)
        return self._sample_with_step(prob)

    def value_to_prob(self, value):
        if self.step is None:
            return self._numerical_to_prob(value)
        return self._to_prob_with_step(value)

    def get_config(self):
        config = super().get_config()
        config["min_value"] = self.min_value
        config["max_value"] = self.max_value
        config["step"] = self.step
        config["sampling"] = self.sampling
        return config

    @classmethod
    def from_proto(cls, proto):
        conditions = [
            conditions_mod.Condition.from_proto(c) for c in proto.conditions
        ]
        return cls(
            name=proto.name,
            min_value=proto.min_value,
            max_value=proto.max_value,
            step=proto.step or None,
            sampling=hp_utils.sampling_from_proto(proto.sampling),
            default=proto.default,
            conditions=conditions,
        )

    def to_proto(self):
        return protos.get_proto().Float(
            name=self.name,
            min_value=self.min_value,
            max_value=self.max_value,
            step=self.step if self.step is not None else 0.0,
            sampling=hp_utils.sampling_to_proto(self.sampling),
            default=self.default,
            conditions=[c.to_proto() for c in self.conditions],
        )




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.engine.objective import Objective as Objective
from keras_tuner.src.engine.oracle import Oracle as Oracle



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from keras_tuner.src.engine.hyperparameters import hp_utils
from keras_tuner.src.engine.hyperparameters import hyperparameter


class Numerical(hyperparameter.HyperParameter):
    """Super class for all numerical type hyperparameters."""

    def __init__(
        self,
        name,
        min_value,
        max_value,
        step=None,
        sampling="linear",
        default=None,
        **kwargs,
    ):
        super().__init__(name=name, default=default, **kwargs)
        self.max_value = max_value
        self.min_value = min_value
        self.step = step
        self.sampling = sampling
        self._check_sampling_arg()

    def _check_sampling_arg(self):
        if self.min_value > self.max_value:
            raise ValueError(
                f"For HyperParameters.{self.__class__.__name__}"
                f"(name='{self.name}'), "
                f"min_value {str(self.min_value)} is greater than "
                f"the max_value {str(self.max_value)}."
            )
        sampling_values = {"linear", "log", "reverse_log"}
        # This is for backward compatibility.
        # sampling=None was allowed and was the default value.
        if self.sampling is None:
            self.sampling = "linear"
        self.sampling = self.sampling.lower()
        if self.sampling not in sampling_values:
            raise ValueError(
                f"For HyperParameters.{self.__class__.__name__}"
                f"(name='{self.name}'), sampling must be one "
                f"of {sampling_values}"
            )

        if self.sampling in {"log", "reverse_log"} and self.min_value <= 0:
            raise ValueError(
                f"For HyperParameters.{self.__class__.__name__}"
                f"(name='{self.name}'), "
                f"sampling='{str(self.sampling)}' does not support "
                f"negative values, found min_value: {str(self.min_value)}."
            )
        if (
            self.sampling in {"log", "reverse_log"}
            and self.step is not None
            and self.step <= 1
        ):
            raise ValueError(
                f"For HyperParameters.{self.__class__.__name__}"
                f"(name='{self.name}'), "
                f"expected step > 1 with sampling='{str(self.sampling)}'. "
                f"Received: step={str(self.step)}."
            )

    def _sample_numerical_value(self, prob, max_value=None):
        """Sample a value with the cumulative prob in the given range."""
        if max_value is None:
            max_value = self.max_value
        if self.sampling == "linear":
            return prob * (max_value - self.min_value) + self.min_value
        elif self.sampling == "log":
            return self.min_value * math.pow(max_value / self.min_value, prob)
        elif self.sampling == "reverse_log":
            return (
                max_value
                + self.min_value
                - self.min_value
                * math.pow(max_value / self.min_value, 1 - prob)
            )

    def _numerical_to_prob(self, value, max_value=None):
        """Convert a numerical value to range [0.0, 1.0)."""
        if max_value is None:
            max_value = self.max_value
        if max_value == self.min_value:
            # Center the prob
            return 0.5
        if self.sampling == "linear":
            return (value - self.min_value) / (max_value - self.min_value)
        if self.sampling == "log":
            return math.log(value / self.min_value) / math.log(
                max_value / self.min_value
            )
        if self.sampling == "reverse_log":
            return 1.0 - math.log(
                (max_value + self.min_value - value) / self.min_value
            ) / math.log(max_value / self.min_value)

    def _get_n_values(self):
        """Get the total number of possible values using step."""
        if self.sampling == "linear":
            # +1 so that max_value may be sampled.
            return int((self.max_value - self.min_value) // self.step + 1)
        # For log and reverse_log
        # +1 so that max_value may be sampled.
        return (
            int(math.log(self.max_value / self.min_value, self.step) + 1e-8) + 1
        )

    def _get_value_by_index(self, index):
        """Get the index-th value in the range given step."""
        if self.sampling == "linear":
            return self.min_value + index * self.step
        if self.sampling == "log":
            return self.min_value * math.pow(self.step, index)
        if self.sampling == "reverse_log":
            return (
                self.max_value
                + self.min_value
                - self.min_value * math.pow(self.step, index)
            )

    def _sample_with_step(self, prob):
        """Sample a value with the cumulative prob in the given range.

        The range is divided evenly by `step`. So only sampling from a finite
        set of values. When calling the function, no need to use (max_value + 1)
        since the function takes care of the inclusion of max_value.
        """
        n_values = self._get_n_values()
        index = hp_utils.prob_to_index(prob, n_values)
        return self._get_value_by_index(index)

    @property
    def values(self):
        if self.step is None:
            # Evenly select 10 samples as the values.
            return tuple(
                {self.prob_to_value(i * 0.1 + 0.05) for i in range(10)}
            )
        n_values = self._get_n_values()
        return (self._get_value_by_index(i) for i in range(n_values))

    def _to_prob_with_step(self, value):
        """Convert to cumulative prob with step specified.

        When calling the function, no need to use (max_value + 1) since the
        function takes care of the inclusion of max_value.
        """
        if self.sampling == "linear":
            index = (value - self.min_value) // self.step
        if self.sampling == "log":
            index = math.log(value / self.min_value, self.step)
        if self.sampling == "reverse_log":
            index = math.log(
                (self.max_value - value + self.min_value) / self.min_value,
                self.step,
            )
        n_values = self._get_n_values()
        return hp_utils.index_to_prob(index, n_values)




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"The Tuner class."


import contextlib
import copy
import gc
import math
import os

import numpy as np

from keras_tuner.src import backend
from keras_tuner.src import errors
from keras_tuner.src import utils
from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.backend import config
from keras_tuner.src.backend import keras
from keras_tuner.src.engine import base_tuner
from keras_tuner.src.engine import tuner_utils


@keras_tuner_export(
    [
        "keras_tuner.Tuner",
        "keras_tuner.tuners.Tuner",
        "keras_tuner.engine.tuner.Tuner",
    ]
)
class Tuner(base_tuner.BaseTuner):
    """Tuner class for Keras models.

    This is the base `Tuner` class for all tuners for Keras models. It manages
    the building, training, evaluation and saving of the Keras models. New
    tuners can be created by subclassing the class.

    All Keras related logics are in `Tuner.run_trial()` and its subroutines.
    When subclassing `Tuner`, if not calling `super().run_trial()`, it can tune
    anything.

    Args:
        oracle: Instance of `Oracle` class.
        hypermodel: Instance of `HyperModel` class (or callable that takes
            hyperparameters and returns a `Model` instance). It is optional
            when `Tuner.run_trial()` is overriden and does not use
            `self.hypermodel`.
        max_model_size: Integer, maximum number of scalars in the parameters of
            a model. Models larger than this are rejected.
        optimizer: Optional optimizer. It is used to override the `optimizer`
            argument in the `compile` step for the models. If the hypermodel
            does not compile the models it generates, then this argument must be
            specified.
        loss: Optional loss. May be used to override the `loss` argument in the
            `compile` step for the models. If the hypermodel does not compile
            the models it generates, then this argument must be specified.
        metrics: Optional metrics. May be used to override the `metrics`
            argument in the `compile` step for the models. If the hypermodel
            does not compile the models it generates, then this argument must
            be specified.
        distribution_strategy: Optional instance of `tf.distribute.Strategy`.
            If specified, each trial will run under this scope. For example,
            `tf.distribute.MirroredStrategy(['/gpu:0', '/gpu:1'])` will run
            each trial on two GPUs. Currently only single-worker strategies are
            supported.
        directory: A string, the relative path to the working directory.
        project_name: A string, the name to use as prefix for files saved by
            this `Tuner`.
        tuner_id: Optional string, used as the ID of this `Tuner`.
        overwrite: Boolean, defaults to `False`. If `False`, reloads an
            existing project of the same name if one is found. Otherwise,
            overwrites the project.
        executions_per_trial: Integer, the number of executions (training a
            model from scratch, starting from a new initialization) to run per
            trial (model configuration). Model metrics may vary greatly
            depending on random initialization, hence it is often a good idea
            to run several executions per trial in order to evaluate the
            performance of a given set of hyperparameter values.
        **kwargs: Arguments for `BaseTuner`.

    Attributes:
        remaining_trials: Number of trials remaining, `None` if `max_trials` is
            not set. This is useful when resuming a previously stopped search.
    """

    def __init__(
        self,
        oracle,
        hypermodel=None,
        max_model_size=None,
        optimizer=None,
        loss=None,
        metrics=None,
        distribution_strategy=None,
        directory=None,
        project_name=None,
        logger=None,
        tuner_id=None,
        overwrite=False,
        executions_per_trial=1,
        **kwargs,
    ):
        if hypermodel is None and self.__class__.run_trial is Tuner.run_trial:
            raise ValueError(
                "Received `hypermodel=None`. We only allow not specifying "
                "`hypermodel` if the user defines the search space in "
                "`Tuner.run_trial()` by subclassing a `Tuner` class without "
                "using a `HyperModel` instance."
            )

        super().__init__(
            oracle=oracle,
            hypermodel=hypermodel,
            directory=directory,
            project_name=project_name,
            logger=logger,
            overwrite=overwrite,
            **kwargs,
        )

        self.max_model_size = max_model_size
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics
        self.distribution_strategy = distribution_strategy
        self.executions_per_trial = executions_per_trial

        # Support multi-worker distribution strategies w/ distributed tuning.
        # Only the chief worker in each cluster should report results.
        if self.distribution_strategy is not None and hasattr(
            self.distribution_strategy.extended, "_in_multi_worker_mode"
        ):
            self.oracle.multi_worker = (
                self.distribution_strategy.extended._in_multi_worker_mode()
            )
            self.oracle.should_report = (
                self.distribution_strategy.extended.should_checkpoint
            )

        self.tuner_id = tuner_id or self.tuner_id

    def _build_hypermodel(self, hp):
        with maybe_distribute(self.distribution_strategy):
            model = self.hypermodel.build(hp)
            self._override_compile_args(model)
            return model

    def _try_build(self, hp):
        # clean-up TF graph from previously stored (defunct) graph
        keras.backend.clear_session()
        gc.collect()

        model = self._build_hypermodel(hp)
        # Stop if `build()` does not return a valid model.
        if not isinstance(model, keras.models.Model):
            raise errors.FatalTypeError(
                "Expected the model-building function, or HyperModel.build() "
                "to return a valid Keras Model instance. "
                f"Received: {model} of type {type(model)}."
            )
        # Check model size.
        size = maybe_compute_model_size(model)
        if self.max_model_size and size > self.max_model_size:
            raise errors.FailedTrialError(
                f"Oversized model: {size} parameters. Skip model."
            )
        return model

    def _filter_metrics(self, metrics):
        if metrics is None:
            return None
        return list(
            filter(
                lambda metric: metric.name not in ("loss", "compile_metric"),
                metrics,
            )
        )

    def _override_compile_args(self, model):
        with maybe_distribute(self.distribution_strategy):
            if self.optimizer or self.loss or self.metrics:
                compile_kwargs = {
                    "optimizer": model.optimizer,
                    "loss": model.loss,
                    "metrics": self._filter_metrics(model.metrics),
                }
                if self.loss:
                    compile_kwargs["loss"] = self.loss
                if self.optimizer:
                    optimizer = (
                        self.optimizer
                        if isinstance(self.optimizer, str)
                        else keras.optimizers.deserialize(
                            keras.optimizers.serialize(self.optimizer)
                        )
                    )
                    compile_kwargs["optimizer"] = optimizer
                if self.metrics:
                    compile_kwargs["metrics"] = self.metrics
                model.compile(**compile_kwargs)

    def _build_and_fit_model(self, trial, *args, **kwargs):
        """For AutoKeras to override.

        DO NOT REMOVE this function. AutoKeras overrides the function to tune
        tf.data preprocessing pipelines, preprocess the dataset to obtain
        the input shape before building the model, adapt preprocessing layers,
        and tune other fit_args and fit_kwargs.

        Args:
            trial: A `Trial` instance that contains the information needed to
                run this trial. `Hyperparameters` can be accessed via
                `trial.hyperparameters`.
            *args: Positional arguments passed by `search`.
            **kwargs: Keyword arguments passed by `search`.

        Returns:
            The fit history.
        """
        hp = trial.hyperparameters
        model = self._try_build(hp)
        results = self.hypermodel.fit(hp, model, *args, **kwargs)

        # Save the build config for model loading later.
        if backend.config.multi_backend():
            utils.save_json(
                self._get_build_config_fname(trial.trial_id),
                model.get_build_config(),
            )

        tuner_utils.validate_trial_results(
            results, self.oracle.objective, "HyperModel.fit()"
        )
        return results

    def run_trial(self, trial, *args, **kwargs):
        """Evaluates a set of hyperparameter values.

        This method is called multiple times during `search` to build and
        evaluate the models with different hyperparameters and return the
        objective value.

        Example:

        You can use it with `self.hypermodel` to build and fit the model.

        ```python
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            model = self.hypermodel.build(hp)
            return self.hypermodel.fit(hp, model, *args, **kwargs)
        ```

        You can also use it as a black-box optimizer for anything.

        ```python
        def run_trial(self, trial, *args, **kwargs):
            hp = trial.hyperparameters
            x = hp.Float("x", -2.0, 2.0)
            y = x * x + 2 * x + 1
            return y
        ```

        Args:
            trial: A `Trial` instance that contains the information needed to
                run this trial. Hyperparameters can be accessed via
                `trial.hyperparameters`.
            *args: Positional arguments passed by `search`.
            **kwargs: Keyword arguments passed by `search`.

        Returns:
            A `History` object, which is the return value of `model.fit()`, a
            dictionary, a float, or a list of one of these types.

            If return a dictionary, it should be a dictionary of the metrics to
            track. The keys are the metric names, which contains the
            `objective` name. The values should be the metric values.

            If return a float, it should be the `objective` value.

            If evaluating the model for multiple times, you may return a list
            of results of any of the types above. The final objective value is
            the average of the results in the list.
        """
        # Not using `ModelCheckpoint` to support MultiObjective.
        # It can only track one of the metrics to save the best model.
        model_checkpoint = tuner_utils.SaveBestEpoch(
            objective=self.oracle.objective,
            filepath=self._get_checkpoint_fname(trial.trial_id),
        )
        original_callbacks = kwargs.pop("callbacks", [])

        # Run the training process multiple times.
        histories = []
        for execution in range(self.executions_per_trial):
            copied_kwargs = copy.copy(kwargs)
            callbacks = self._deepcopy_callbacks(original_callbacks)
            self._configure_tensorboard_dir(callbacks, trial, execution)
            callbacks.append(tuner_utils.TunerCallback(self, trial))
            # Only checkpoint the best epoch across all executions.
            callbacks.append(model_checkpoint)
            copied_kwargs["callbacks"] = callbacks
            obj_value = self._build_and_fit_model(trial, *args, **copied_kwargs)

            histories.append(obj_value)
        return histories

    def load_model(self, trial):
        model = self._try_build(trial.hyperparameters)

        # Build model to create the weights.
        if backend.config.multi_backend() and not model.built:
            model.build_from_config(
                utils.load_json(self._get_build_config_fname(trial.trial_id))
            )

        # Reload best checkpoint.
        # Only load weights to avoid loading `custom_objects`.
        with maybe_distribute(self.distribution_strategy):
            model.load_weights(self._get_checkpoint_fname(trial.trial_id))
        return model

    def on_batch_begin(self, trial, model, batch, logs):
        """Called at the beginning of a batch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            batch: The current batch number within the current epoch.
            logs: Additional metrics.
        """
        pass

    def on_batch_end(self, trial, model, batch, logs=None):
        """Called at the end of a batch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            batch: The current batch number within the current epoch.
            logs: Additional metrics.
        """
        pass

    def on_epoch_begin(self, trial, model, epoch, logs=None):
        """Called at the beginning of an epoch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            epoch: The current epoch number.
            logs: Additional metrics.
        """
        pass

    def on_epoch_end(self, trial, model, epoch, logs=None):
        """Called at the end of an epoch.

        Args:
            trial: A `Trial` instance.
            model: A Keras `Model`.
            epoch: The current epoch number.
            logs: Dict. Metrics for this epoch. This should include
              the value of the objective for this epoch.
        """
        # Intermediate results are not passed to the Oracle, and
        # checkpointing is handled via a `SaveBestEpoch` callback.
        pass

    def get_best_models(self, num_models=1):
        """Returns the best model(s), as determined by the tuner's objective.

        The models are loaded with the weights corresponding to
        their best checkpoint (at the end of the best epoch of best trial).

        This method is for querying the models trained during the search.
        For best performance, it is recommended to retrain your Model on the
        full dataset using the best hyperparameters found during `search`,
        which can be obtained using `tuner.get_best_hyperparameters()`.

        Args:
            num_models: Optional number of best models to return.
                Defaults to 1.

        Returns:
            List of trained model instances sorted from the best to the worst.
        """
        # Method only exists in this class for the docstring override.
        return super().get_best_models(num_models)

    def _deepcopy_callbacks(self, callbacks):
        try:
            callbacks = copy.deepcopy(callbacks)
        except:
            raise errors.FatalValueError(
                "All callbacks used during a search "
                "should be deep-copyable (since they are "
                "reused across trials). "
                "It is not possible to do `copy.deepcopy(%s)`" % (callbacks,)
            )
        return callbacks

    def _configure_tensorboard_dir(self, callbacks, trial, execution=0):
        # Only import tensorboard when using tensorflow backend to avoid
        # importing tensorflow with other backend (tensorboard would import
        # tensorflow).
        if backend.config.backend() != "tensorflow":
            return

        from tensorboard.plugins.hparams import api as hparams_api

        for callback in callbacks:
            if callback.__class__.__name__ == "TensorBoard":
                # Patch TensorBoard log_dir and add HParams KerasCallback
                logdir = self._get_tensorboard_dir(
                    callback.log_dir, trial.trial_id, execution
                )
                callback.log_dir = logdir
                hparams = tuner_utils.convert_hyperparams_to_hparams(
                    trial.hyperparameters,
                    hparams_api,
                )
                callbacks.append(
                    hparams_api.KerasCallback(
                        writer=logdir, hparams=hparams, trial_id=trial.trial_id
                    )
                )

    def _get_tensorboard_dir(self, logdir, trial_id, execution):
        return os.path.join(
            str(logdir), str(trial_id), f"execution{str(execution)}"
        )

    def _get_checkpoint_fname(self, trial_id):
        return os.path.join(
            # Each checkpoint is saved in its own directory.
            self.get_trial_dir(trial_id),
            "checkpoint.weights.h5" if config.multi_backend() else "checkpoint",
        )

    def _get_build_config_fname(self, trial_id):
        return os.path.join(
            # Each checkpoint is saved in its own directory.
            self.get_trial_dir(trial_id),
            "build_config.json",
        )


def maybe_compute_model_size(model):
    """Compute the size of a given model, if it has been built."""
    if model.built:
        params = [math.prod(p.shape) for p in model.trainable_weights]
        return int(np.sum(params))
    return 0


@keras_tuner_export("keras_tuner.engine.tuner.maybe_distribute")
@contextlib.contextmanager
def maybe_distribute(distribution_strategy):
    """Distributes if distribution_strategy is set."""
    if distribution_strategy is None:
        yield
    else:
        with distribution_strategy.scope():
            yield




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hypertunable version of ResNet."""

from keras_tuner.src.api_export import keras_tuner_export
from keras_tuner.src.backend import keras
from keras_tuner.src.backend import ops
from keras_tuner.src.backend.keras import layers
from keras_tuner.src.engine import hypermodel


@keras_tuner_export("keras_tuner.applications.HyperResNet")
class HyperResNet(hypermodel.HyperModel):
    """A ResNet hypermodel.

    Models built by `HyperResNet` take images with shape (height, width,
    channels) as input. The output are one-hot encoded with the length matching
    the number of classes specified by the `classes` argument.

    Args:
        include_top: Boolean, whether to include the fully-connected layer at
            the top of the network.
        input_shape: Optional shape tuple, e.g. `(256, 256, 3)`.  One of
            `input_shape` or `input_tensor` must be specified.
        input_tensor: Optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.  One of `input_shape` or
            `input_tensor` must be specified.
        classes: Optional number of classes to classify images into, only to be
            specified if `include_top` is True, and if no `weights` argument is
            specified.
        **kwargs: Additional keyword arguments that apply to all hypermodels.
            See `keras_tuner.HyperModel`.
    """

    def __init__(
        self,
        include_top=True,
        input_shape=None,
        input_tensor=None,
        classes=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if include_top and classes is None:
            raise ValueError(
                "You must specify `classes` when `include_top=True`"
            )

        if input_shape is None and input_tensor is None:
            raise ValueError(
                "You must specify either `input_shape` or `input_tensor`."
            )

        self.include_top = include_top
        self.input_shape = input_shape
        self.input_tensor = input_tensor
        self.classes = classes

    def build(self, hp):
        version = hp.Choice("version", ["v1", "v2", "next"], default="v2")
        conv3_depth = hp.Choice("conv3_depth", [4, 8])
        conv4_depth = hp.Choice("conv4_depth", [6, 23, 36])

        # Version-conditional fixed parameters
        preact = version == "v2"
        use_bias = version != "next"

        # Model definition.
        bn_axis = (
            3 if keras.backend.image_data_format() == "channels_last" else 1
        )

        if self.input_tensor is not None:
            inputs = keras.utils.get_source_inputs(self.input_tensor)
            x = self.input_tensor
        else:
            inputs = layers.Input(shape=self.input_shape)
            x = inputs

        # Initial conv2d block.
        x = layers.ZeroPadding2D(padding=((3, 3), (3, 3)), name="conv1_pad")(x)
        x = layers.Conv2D(
            64, 7, strides=2, use_bias=use_bias, name="conv1_conv"
        )(x)
        if not preact:
            x = layers.BatchNormalization(
                axis=bn_axis, epsilon=1.001e-5, name="conv1_bn"
            )(x)
            x = layers.Activation("relu", name="conv1_relu")(x)
        x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name="pool1_pad")(x)
        x = layers.MaxPooling2D(3, strides=2, name="pool1_pool")(x)

        # Middle hypertunable stack.
        if version == "v1":
            x = stack1(x, 64, 3, stride1=1, name="conv2")
            x = stack1(x, 128, conv3_depth, name="conv3")
            x = stack1(x, 256, conv4_depth, name="conv4")
            x = stack1(x, 512, 3, name="conv5")
        elif version == "v2":
            x = stack2(x, 64, 3, name="conv2")
            x = stack2(x, 128, conv3_depth, name="conv3")
            x = stack2(x, 256, conv4_depth, name="conv4")
            x = stack2(x, 512, 3, stride1=1, name="conv5")
        elif version == "next":
            x = stack3(x, 64, 3, name="conv2")
            x = stack3(x, 256, conv3_depth, name="conv3")
            x = stack3(x, 512, conv4_depth, name="conv4")
            x = stack3(x, 1024, 3, stride1=1, name="conv5")

        # Top of the model.
        if preact:
            x = layers.BatchNormalization(
                axis=bn_axis, epsilon=1.001e-5, name="post_bn"
            )(x)
            x = layers.Activation("relu", name="post_relu")(x)

        pooling = hp.Choice("pooling", ["avg", "max"], default="avg")
        if pooling == "avg":
            x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
        elif pooling == "max":
            x = layers.GlobalMaxPooling2D(name="max_pool")(x)

        if not self.include_top:
            return keras.Model(inputs, x, name="ResNet")
        x = layers.Dense(self.classes, activation="softmax", name="probs")(x)
        model = keras.Model(inputs, x, name="ResNet")
        optimizer_name = hp.Choice(
            "optimizer", ["adam", "rmsprop", "sgd"], default="adam"
        )
        optimizer = keras.optimizers.get(optimizer_name)
        optimizer.learning_rate = hp.Choice(
            "learning_rate", [0.1, 0.01, 0.001], default=0.01
        )
        model.compile(
            optimizer=optimizer,
            loss="categorical_crossentropy",
            metrics=["accuracy"],
        )
        return model


def block1(x, filters, kernel_size=3, stride=1, conv_shortcut=True, name=None):
    """A residual block.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        conv_shortcut: default True, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    Returns:
        Output tensor for the residual block.
    """
    bn_axis = 3 if keras.backend.image_data_format() == "channels_last" else 1

    if conv_shortcut is True:
        shortcut = layers.Conv2D(
            4 * filters, 1, strides=stride, name=f"{name}_0_conv"
        )(x)

        shortcut = layers.BatchNormalization(
            axis=bn_axis, epsilon=1.001e-5, name=f"{name}_0_bn"
        )(shortcut)

    else:
        shortcut = x

    x = layers.Conv2D(filters, 1, strides=stride, name=f"{name}_1_conv")(x)
    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_1_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_1_relu")(x)

    x = layers.Conv2D(
        filters, kernel_size, padding="same", name=f"{name}_2_conv"
    )(x)

    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_2_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_2_relu")(x)

    x = layers.Conv2D(4 * filters, 1, name=f"{name}_3_conv")(x)
    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_3_bn"
    )(x)

    x = layers.Add(name=f"{name}_add")([shortcut, x])
    x = layers.Activation("relu", name=f"{name}_out")(x)
    return x


def stack1(x, filters, blocks, stride1=2, name=None):
    """A set of stacked residual blocks.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        name: string, stack label.

    Returns:
        Output tensor for the stacked blocks.
    """
    x = block1(x, filters, stride=stride1, name=f"{name}_block1")
    for i in range(2, blocks + 1):
        x = block1(
            x, filters, conv_shortcut=False, name=f"{name}_block{str(i)}"
        )
    return x


def block2(x, filters, kernel_size=3, stride=1, conv_shortcut=False, name=None):
    """A residual block.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        conv_shortcut: default False, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    Returns:
        Output tensor for the residual block.
    """
    bn_axis = 3 if keras.backend.image_data_format() == "channels_last" else 1

    preact = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_preact_bn"
    )(x)

    preact = layers.Activation("relu", name=f"{name}_preact_relu")(preact)

    if conv_shortcut is True:
        shortcut = layers.Conv2D(
            4 * filters, 1, strides=stride, name=f"{name}_0_conv"
        )(preact)

    else:
        shortcut = (
            layers.MaxPooling2D(1, strides=stride)(x) if stride > 1 else x
        )

    x = layers.Conv2D(
        filters, 1, strides=1, use_bias=False, name=f"{name}_1_conv"
    )(preact)

    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_1_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_1_relu")(x)

    x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name=f"{name}_2_pad")(x)
    x = layers.Conv2D(
        filters,
        kernel_size,
        strides=stride,
        use_bias=False,
        name=f"{name}_2_conv",
    )(x)

    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_2_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_2_relu")(x)

    x = layers.Conv2D(4 * filters, 1, name=f"{name}_3_conv")(x)
    x = layers.Add(name=f"{name}_out")([shortcut, x])
    return x


def stack2(x, filters, blocks, stride1=2, name=None):
    """A set of stacked residual blocks.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        name: string, stack label.

    Returns:
        Output tensor for the stacked blocks.
    """
    x = block2(x, filters, conv_shortcut=True, name=f"{name}_block1")
    for i in range(2, blocks):
        x = block2(x, filters, name=f"{name}_block{str(i)}")
    x = block2(x, filters, stride=stride1, name=f"{name}_block{str(blocks)}")
    return x


def block3(
    x,
    filters,
    kernel_size=3,
    stride=1,
    groups=32,
    conv_shortcut=True,
    name=None,
):
    """A residual block.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer.
        kernel_size: default 3, kernel size of the bottleneck layer.
        stride: default 1, stride of the first layer.
        groups: default 32, group size for grouped convolution.
        conv_shortcut: default True, use convolution shortcut if True,
            otherwise identity shortcut.
        name: string, block label.

    Returns:
        Output tensor for the residual block.
    """
    bn_axis = 3 if keras.backend.image_data_format() == "channels_last" else 1

    if conv_shortcut is True:
        shortcut = layers.Conv2D(
            (64 // groups) * filters,
            1,
            strides=stride,
            use_bias=False,
            name=f"{name}_0_conv",
        )(x)

        shortcut = layers.BatchNormalization(
            axis=bn_axis, epsilon=1.001e-5, name=f"{name}_0_bn"
        )(shortcut)

    else:
        shortcut = x

    x = layers.Conv2D(filters, 1, use_bias=False, name=f"{name}_1_conv")(x)
    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_1_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_1_relu")(x)

    c = filters // groups
    x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name=f"{name}_2_pad")(x)
    x = layers.DepthwiseConv2D(
        kernel_size,
        strides=stride,
        depth_multiplier=c,
        use_bias=False,
        name=f"{name}_2_conv",
    )(x)

    if bn_axis == 3:
        x_shape = ops.shape(x)[1:-1]
        x = layers.Reshape(x_shape + (groups, c, c))(x)
        output_shape = x_shape + (groups, c)
        x = layers.Lambda(
            lambda x: sum(x[:, :, :, :, i] for i in range(c)),
            name=f"{name}_2_reduce",
            output_shape=output_shape,
        )(x)

        x = layers.Reshape(x_shape + (filters,))(x)
    else:
        x_shape = ops.shape(x)[2:]
        x = layers.Reshape((groups, c, c) + x_shape)(x)
        output_shape = (groups, c) + x_shape
        x = layers.Lambda(
            lambda x: sum(x[:, :, i, :, :] for i in range(c)),
            name=f"{name}_2_reduce",
            output_shape=output_shape,
        )(x)

        x = layers.Reshape((filters,) + x_shape)(x)

    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_2_bn"
    )(x)

    x = layers.Activation("relu", name=f"{name}_2_relu")(x)

    x = layers.Conv2D(
        (64 // groups) * filters, 1, use_bias=False, name=f"{name}_3_conv"
    )(x)

    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=f"{name}_3_bn"
    )(x)

    x = layers.Add(name=f"{name}_add")([shortcut, x])
    x = layers.Activation("relu", name=f"{name}_out")(x)
    return x


def stack3(x, filters, blocks, stride1=2, groups=32, name=None):
    """A set of stacked residual blocks.

    Args:
        x: input tensor.
        filters: integer, filters of the bottleneck layer in a block.
        blocks: integer, blocks in the stacked blocks.
        stride1: default 2, stride of the first layer in the first block.
        groups: default 32, group size for grouped convolution.
        name: string, stack label.

    Returns:
        Output tensor for the stacked blocks.
    """
    x = block3(x, filters, stride=stride1, groups=groups, name=f"{name}_block1")

    for i in range(2, blocks + 1):
        x = block3(
            x,
            filters,
            groups=groups,
            conv_shortcut=False,
            name=f"{name}_block{str(i)}",
        )

    return x




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keras_tuner/protos/service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from keras_tuner.src.protos.v3 import (
    keras_tuner_pb2 as keras__tuner_dot_protos_dot_keras__tuner__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n keras_tuner/protos/service.proto\x12\x0bkeras_tuner\x1a$keras_tuner/protos/keras_tuner.proto"\x11\n\x0fGetSpaceRequest"I\n\x10GetSpaceResponse\x12\x35\n\x0fhyperparameters\x18\x01 \x01(\x0b\x32\x1c.keras_tuner.HyperParameters"K\n\x12UpdateSpaceRequest\x12\x35\n\x0fhyperparameters\x18\x01 \x01(\x0b\x32\x1c.keras_tuner.HyperParameters"\x15\n\x13UpdateSpaceResponse"&\n\x12\x43reateTrialRequest\x12\x10\n\x08tuner_id\x18\x01 \x01(\t"8\n\x13\x43reateTrialResponse\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial"\xa3\x01\n\x12UpdateTrialRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12=\n\x07metrics\x18\x02 \x03(\x0b\x32,.keras_tuner.UpdateTrialRequest.MetricsEntry\x12\x0c\n\x04step\x18\x03 \x01(\x03\x1a.\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01"8\n\x13UpdateTrialResponse\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial"4\n\x0f\x45ndTrialRequest\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial"\x12\n\x10\x45ndTrialResponse"*\n\x14GetBestTrialsRequest\x12\x12\n\nnum_trials\x18\x01 \x01(\x03";\n\x15GetBestTrialsResponse\x12"\n\x06trials\x18\x01 \x03(\x0b\x32\x12.keras_tuner.Trial"#\n\x0fGetTrialRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t"5\n\x10GetTrialResponse\x12!\n\x05trial\x18\x01 \x01(\x0b\x32\x12.keras_tuner.Trial2\xbf\x04\n\x06Oracle\x12I\n\x08GetSpace\x12\x1c.keras_tuner.GetSpaceRequest\x1a\x1d.keras_tuner.GetSpaceResponse"\x00\x12R\n\x0bUpdateSpace\x12\x1f.keras_tuner.UpdateSpaceRequest\x1a .keras_tuner.UpdateSpaceResponse"\x00\x12R\n\x0b\x43reateTrial\x12\x1f.keras_tuner.CreateTrialRequest\x1a .keras_tuner.CreateTrialResponse"\x00\x12R\n\x0bUpdateTrial\x12\x1f.keras_tuner.UpdateTrialRequest\x1a .keras_tuner.UpdateTrialResponse"\x00\x12I\n\x08\x45ndTrial\x12\x1c.keras_tuner.EndTrialRequest\x1a\x1d.keras_tuner.EndTrialResponse"\x00\x12X\n\rGetBestTrials\x12!.keras_tuner.GetBestTrialsRequest\x1a".keras_tuner.GetBestTrialsResponse"\x00\x12I\n\x08GetTrial\x12\x1c.keras_tuner.GetTrialRequest\x1a\x1d.keras_tuner.GetTrialResponse"\x00\x62\x06proto3'
)


_GETSPACEREQUEST = DESCRIPTOR.message_types_by_name["GetSpaceRequest"]
_GETSPACERESPONSE = DESCRIPTOR.message_types_by_name["GetSpaceResponse"]
_UPDATESPACEREQUEST = DESCRIPTOR.message_types_by_name["UpdateSpaceRequest"]
_UPDATESPACERESPONSE = DESCRIPTOR.message_types_by_name["UpdateSpaceResponse"]
_CREATETRIALREQUEST = DESCRIPTOR.message_types_by_name["CreateTrialRequest"]
_CREATETRIALRESPONSE = DESCRIPTOR.message_types_by_name["CreateTrialResponse"]
_UPDATETRIALREQUEST = DESCRIPTOR.message_types_by_name["UpdateTrialRequest"]
_UPDATETRIALREQUEST_METRICSENTRY = _UPDATETRIALREQUEST.nested_types_by_name[
    "MetricsEntry"
]
_UPDATETRIALRESPONSE = DESCRIPTOR.message_types_by_name["UpdateTrialResponse"]
_ENDTRIALREQUEST = DESCRIPTOR.message_types_by_name["EndTrialRequest"]
_ENDTRIALRESPONSE = DESCRIPTOR.message_types_by_name["EndTrialResponse"]
_GETBESTTRIALSREQUEST = DESCRIPTOR.message_types_by_name["GetBestTrialsRequest"]
_GETBESTTRIALSRESPONSE = DESCRIPTOR.message_types_by_name[
    "GetBestTrialsResponse"
]
_GETTRIALREQUEST = DESCRIPTOR.message_types_by_name["GetTrialRequest"]
_GETTRIALRESPONSE = DESCRIPTOR.message_types_by_name["GetTrialResponse"]
GetSpaceRequest = _reflection.GeneratedProtocolMessageType(
    "GetSpaceRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSPACEREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetSpaceRequest)
    },
)
_sym_db.RegisterMessage(GetSpaceRequest)

GetSpaceResponse = _reflection.GeneratedProtocolMessageType(
    "GetSpaceResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSPACERESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetSpaceResponse)
    },
)
_sym_db.RegisterMessage(GetSpaceResponse)

UpdateSpaceRequest = _reflection.GeneratedProtocolMessageType(
    "UpdateSpaceRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATESPACEREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.UpdateSpaceRequest)
    },
)
_sym_db.RegisterMessage(UpdateSpaceRequest)

UpdateSpaceResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateSpaceResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATESPACERESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.UpdateSpaceResponse)
    },
)
_sym_db.RegisterMessage(UpdateSpaceResponse)

CreateTrialRequest = _reflection.GeneratedProtocolMessageType(
    "CreateTrialRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATETRIALREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.CreateTrialRequest)
    },
)
_sym_db.RegisterMessage(CreateTrialRequest)

CreateTrialResponse = _reflection.GeneratedProtocolMessageType(
    "CreateTrialResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATETRIALRESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.CreateTrialResponse)
    },
)
_sym_db.RegisterMessage(CreateTrialResponse)

UpdateTrialRequest = _reflection.GeneratedProtocolMessageType(
    "UpdateTrialRequest",
    (_message.Message,),
    {
        "MetricsEntry": _reflection.GeneratedProtocolMessageType(
            "MetricsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _UPDATETRIALREQUEST_METRICSENTRY,
                "__module__": "keras_tuner.protos.service_pb2",
                # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialRequest.MetricsEntry)
            },
        ),
        "DESCRIPTOR": _UPDATETRIALREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialRequest)
    },
)
_sym_db.RegisterMessage(UpdateTrialRequest)
_sym_db.RegisterMessage(UpdateTrialRequest.MetricsEntry)

UpdateTrialResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateTrialResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATETRIALRESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.UpdateTrialResponse)
    },
)
_sym_db.RegisterMessage(UpdateTrialResponse)

EndTrialRequest = _reflection.GeneratedProtocolMessageType(
    "EndTrialRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENDTRIALREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.EndTrialRequest)
    },
)
_sym_db.RegisterMessage(EndTrialRequest)

EndTrialResponse = _reflection.GeneratedProtocolMessageType(
    "EndTrialResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENDTRIALRESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.EndTrialResponse)
    },
)
_sym_db.RegisterMessage(EndTrialResponse)

GetBestTrialsRequest = _reflection.GeneratedProtocolMessageType(
    "GetBestTrialsRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETBESTTRIALSREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetBestTrialsRequest)
    },
)
_sym_db.RegisterMessage(GetBestTrialsRequest)

GetBestTrialsResponse = _reflection.GeneratedProtocolMessageType(
    "GetBestTrialsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETBESTTRIALSRESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetBestTrialsResponse)
    },
)
_sym_db.RegisterMessage(GetBestTrialsResponse)

GetTrialRequest = _reflection.GeneratedProtocolMessageType(
    "GetTrialRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETTRIALREQUEST,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetTrialRequest)
    },
)
_sym_db.RegisterMessage(GetTrialRequest)

GetTrialResponse = _reflection.GeneratedProtocolMessageType(
    "GetTrialResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETTRIALRESPONSE,
        "__module__": "keras_tuner.protos.service_pb2",
        # @@protoc_insertion_point(class_scope:keras_tuner.GetTrialResponse)
    },
)
_sym_db.RegisterMessage(GetTrialResponse)

_ORACLE = DESCRIPTOR.services_by_name["Oracle"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _UPDATETRIALREQUEST_METRICSENTRY._options = None
    _UPDATETRIALREQUEST_METRICSENTRY._serialized_options = b"8\001"
    _GETSPACEREQUEST._serialized_start = 87
    _GETSPACEREQUEST._serialized_end = 104
    _GETSPACERESPONSE._serialized_start = 106
    _GETSPACERESPONSE._serialized_end = 179
    _UPDATESPACEREQUEST._serialized_start = 181
    _UPDATESPACEREQUEST._serialized_end = 256
    _UPDATESPACERESPONSE._serialized_start = 258
    _UPDATESPACERESPONSE._serialized_end = 279
    _CREATETRIALREQUEST._serialized_start = 281
    _CREATETRIALREQUEST._serialized_end = 319
    _CREATETRIALRESPONSE._serialized_start = 321
    _CREATETRIALRESPONSE._serialized_end = 377
    _UPDATETRIALREQUEST._serialized_start = 380
    _UPDATETRIALREQUEST._serialized_end = 543
    _UPDATETRIALREQUEST_METRICSENTRY._serialized_start = 497
    _UPDATETRIALREQUEST_METRICSENTRY._serialized_end = 543
    _UPDATETRIALRESPONSE._serialized_start = 545
    _UPDATETRIALRESPONSE._serialized_end = 601
    _ENDTRIALREQUEST._serialized_start = 603
    _ENDTRIALREQUEST._serialized_end = 655
    _ENDTRIALRESPONSE._serialized_start = 657
    _ENDTRIALRESPONSE._serialized_end = 675
    _GETBESTTRIALSREQUEST._serialized_start = 677
    _GETBESTTRIALSREQUEST._serialized_end = 719
    _GETBESTTRIALSRESPONSE._serialized_start = 721
    _GETBESTTRIALSRESPONSE._serialized_end = 780
    _GETTRIALREQUEST._serialized_start = 782
    _GETTRIALREQUEST._serialized_end = 817
    _GETTRIALRESPONSE._serialized_start = 819
    _GETTRIALRESPONSE._serialized_end = 872
    _ORACLE._serialized_start = 875
    _ORACLE._serialized_end = 1450
# @@protoc_insertion_point(module_scope)




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_tuner.src import utils
from keras_tuner.src.backend import keras
from keras_tuner.src.engine.hyperparameters.hp_types.boolean_hp import Boolean
from keras_tuner.src.engine.hyperparameters.hp_types.choice_hp import Choice
from keras_tuner.src.engine.hyperparameters.hp_types.fixed_hp import Fixed
from keras_tuner.src.engine.hyperparameters.hp_types.float_hp import Float
from keras_tuner.src.engine.hyperparameters.hp_types.int_hp import Int

OBJECTS = (
    Fixed,
    Float,
    Int,
    Choice,
    Boolean,
)

ALL_CLASSES = {cls.__name__: cls for cls in OBJECTS}


def deserialize(config):
    return utils.deserialize_keras_object(config, module_objects=ALL_CLASSES)


def serialize(obj):
    return utils.serialize_keras_object(obj)




# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types

from keras_tuner.src.backend.config import multi_backend

if multi_backend():
    from keras import *  # noqa: F403, F401
else:
    import tensorflow as tf
    from tensorflow.keras import *  # noqa: F403, F401

    # Shims to handle symbol renames for older `tf.keras` versions.
    if not hasattr(tf.keras, "saving"):
        saving = types.SimpleNamespace()
    else:
        from tensorflow.keras import saving
    from tensorflow.keras import utils

    if not hasattr(saving, "deserialize_keras_object"):
        saving.deserialize_keras_object = utils.deserialize_keras_object
    if not hasattr(saving, "serialize_keras_object"):
        saving.serialize_keras_object = utils.serialize_keras_object
    if not hasattr(saving, "register_keras_serializable"):
        saving.register_keras_serializable = utils.register_keras_serializable




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner.src.applications.augment import HyperImageAugment as HyperImageAugment
from keras_tuner.src.applications.efficientnet import HyperEfficientNet as HyperEfficientNet
from keras_tuner.src.applications.resnet import HyperResNet as HyperResNet
from keras_tuner.src.applications.xception import HyperXception as HyperXception



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_tuner import applications as applications
from keras_tuner import engine as engine
from keras_tuner import errors as errors
from keras_tuner import oracles as oracles
from keras_tuner import tuners as tuners
from keras_tuner.src.engine.hypermodel import HyperModel as HyperModel
from keras_tuner.src.engine.hyperparameters.hyperparameters import HyperParameters as HyperParameters
from keras_tuner.src.engine.objective import Objective as Objective
from keras_tuner.src.engine.oracle import Oracle as Oracle
from keras_tuner.src.engine.oracle import synchronized as synchronized
from keras_tuner.src.engine.tuner import Tuner as Tuner
from keras_tuner.src.tuners.bayesian import BayesianOptimization as BayesianOptimization
from keras_tuner.src.tuners.gridsearch import GridSearch as GridSearch
from keras_tuner.src.tuners.hyperband import Hyperband as Hyperband
from keras_tuner.src.tuners.randomsearch import RandomSearch as RandomSearch
from keras_tuner.src.tuners.sklearn_tuner import SklearnTuner as SklearnTuner


__version__ = "1.4.8"



# Copyright 2019 The KerasTuner Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Tuner base class."


from keras_tuner.src import utils


class Stateful:
    """The base class for saving and restoring the state.

    The functionalities in this class is for the user to resume a previously
    stopped program by runing the same code again.

    Usually, the arguments in `__init__()` should not be added to the state. The
    `Stateful` objects can get the arguments again just by running the same code
    even without loading the previous states.

    The state of an object here is the results produced during running the
    program, which can save time for the user if restored.
    """

    def get_state(self):
        """Returns the current state of this object.

        This method is called during `save`.

        Returns:
            A dictionary of serializable objects as the state.
        """
        raise NotImplementedError

    def set_state(self, state):
        """Sets the current state of this object.

        This method is called during `reload`.

        Args:
            state: A dictionary of serialized objects as the state to restore.
        """
        raise NotImplementedError

    def save(self, fname):
        """Saves this object using `get_state`.

        Args:
            fname: A string, the file name to save to.

        Returns:
            String. The serialized state of the object.
        """
        return utils.save_json(fname, self.get_state())

    def reload(self, fname):
        """Reloads this object using `set_state`.

        Args:
            fname: A string, the file name to restore from.
        """
        self.set_state(utils.load_json(fname))

